__all__ = ('AssertionRaising',)

from ..helpers.exception_matching import try_match_exception
from ..helpers.un_nesting import un_nest_exceptions

from .assertion_base import AssertionBase
from .assertion_states import ASSERTION_STATE_CREATED, ASSERTION_STATE_FAILED, ASSERTION_STATE_PASSED
from .exceptions import AssertionException


class AssertionRaising(AssertionBase):
    """
    Context manager, which checks for exception raise.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised within the context block if any.
    accept_subtypes : `bool`
        Whether subclasses are accepted as well.
    expected_exceptions : `set` of (`BaseException`, `type<BaseException>`)
        The expected exception types.
    where : `None`, `callable`
        Additional check to check the raised exception.
    
    Examples
    --------
    ```py
    with AssertionRaising(TypeError):
        'nice' + 69
    
    with AssertionRaising(ValueError(1)):
        raise ValueError(1)
    
    with AssertionRaising(ValueError, where = lambda err: err.args == (1,)):
        raise ValueError(1)
    ```
    """
    __slots__ = ('accept_subtypes', 'exception', 'expected_exceptions', 'where')
    
    def __new__(cls, expected_exception, *expected_exceptions, accept_subtypes = True, where = None):
        """
        Creates a new raise asserting context manager.
        
        Parameters
        ----------
        expected_exception : `BaseException˙
            The expected exception to be raised.
        *expected_exceptions : ˙tuple<BaseException>˙
            Additional expected exceptions.
        accept_subtypes : `bool` = `True`, Optional (Keyword only)
            Whether subclasses are accepted as well.
        where : `None`, `callable` = `None`, Optional (Keyword only)
            Additional check to check the raised exception.
        
        Raises
        ------
        TypeError
            If an `expected_exceptions`'s type is incorrect.
        ValueError
            If no exception was passed.
        """
        expected_exceptions = un_nest_exceptions((expected_exception, *expected_exceptions))
        if not expected_exceptions:
            raise ValueError(
                'At least 1 exception is required.'
            )
        
        self = AssertionBase.__new__(cls)
        self.accept_subtypes = accept_subtypes
        self.expected_exceptions = expected_exceptions
        self.where = where
        
        return self
    
    
    def __enter__(self):
        """Enters as context manager."""
        self.state = ASSERTION_STATE_CREATED
        return None
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """Exits the context manager."""
        if exception_type is None:
            self.state = ASSERTION_STATE_FAILED
            try:
                raise AssertionException(self)
            finally:
                # Clear reference to self
                self = None
        
        self.exception = exception_value
        
        if try_match_exception(self.expected_exceptions, exception_value, self.accept_subtypes, self.where):
            state = ASSERTION_STATE_PASSED
            silence = True
        
        else:
            state = ASSERTION_STATE_FAILED
            silence = False
        
        self.state = state
        return silence



__all__ = ('get_function_environment_identifier', 'shutdown_environments',)

from scarletio import is_coroutine_function,is_coroutine_generator_function, is_generator_function

from .configuration import ENVIRONMENTS_BY_SCOPE
from .constants import (
    ENVIRONMENT_TYPE_COROUTINE, ENVIRONMENT_TYPE_COROUTINE_GENERATOR, ENVIRONMENT_TYPE_DEFAULT,
    ENVIRONMENT_TYPE_GENERATOR
)


def get_function_environment_identifier(func):
    """
    Gets the environment type depending on the given function's type.
    
    Parameters
    ----------
    func : `FunctionType`
        The function to check.
    
    Returns
    -------
    identifier : `int`
    """
    if is_coroutine_generator_function(func):
        return ENVIRONMENT_TYPE_COROUTINE_GENERATOR
    
    if is_coroutine_function(func):
        return ENVIRONMENT_TYPE_COROUTINE
    
    if is_generator_function(func):
        return ENVIRONMENT_TYPE_GENERATOR
    
    return ENVIRONMENT_TYPE_DEFAULT


def shutdown_environments():
    """
    Shuts down all the registered environments. This operation cannot be reversed.
    """
    for environment_by_detail in ENVIRONMENTS_BY_SCOPE.values():
        for environments_by_identifier in environment_by_detail.values():
            for environment in environments_by_identifier.values():
                environment.shutdown()



__all__ = ()

import sys
from os.path import join as join_paths, isfile as is_file
from types import ModuleType
from importlib.util import module_from_spec
from importlib.machinery import ModuleSpec, SourceFileLoader
from os import getcwd as get_current_working_directory, sep as PATH_SEPARATOR
from os.path import dirname as get_parent_directory_path, split as split_path, splitext as split_extension


PYTHON_EXTENSIONS = frozenset(('.py', '.pyd', '.pyc', '.so'))


def has_init(path):
    """
    Returns whether the directory has `__init__.py` file, or something equivalent to it.
    
    Parameters
    ----------
    path : `str`
        The path to check.
    
    Returns
    -------
    has_init : `bool`
    """
    path = join_paths(path, '__init__')
    for python_extension in PYTHON_EXTENSIONS:
        if is_file(path + python_extension):
            return True
    
    return False


def is_directory_project_source(path):
    """
    Returns whether the given directory is a project source.
    
    Parameters
    ----------
    path : `str`
        The path to check.
    
    Returns
    -------
    is_directory_project_source : `bool`
    """
    if is_file(join_paths(path, 'setup.py')):
        return True
    
    if is_file(join_paths(path, 'pyproject.toml')):
        return True
    
    return False


def build_project_sources(module_name, module_packages):
    """
    Builds the project's sources to import.
    
    Parameters
    ----------
    module_name : `object`
        The module's name. Preferably a string.
    module_packages : `object`
        The module's packages. Preferably a list of strings.
    
    Returns
    -------
    project_sources : `None | set<str>`
    """
    project_sources = None
    
    # check `module_packages` if applicable
    if (module_packages is not None) and (getattr(module_packages, '__iter__', None) is not None):
        for module_package in module_packages:
            if not isinstance(module_package, str):
                continue
            
            dot_index = module_package.find('.')
            if dot_index == -1:
                project_source = module_package
            elif dot_index > 0:
                project_source = module_package[:dot_index]
            else:
                continue
            
            if project_sources is None:
                project_sources = set()
            
            project_sources.add(project_source)
    
    # check `module_name` if applicable
    elif (module_name is not None) and isinstance(module_name, str):
        if project_sources is None:
            project_sources = set()
        
        project_sources.add(module_name)
    
    return project_sources


# I ain't gonna write test for this
def execute_setup_file(path):
    """
    Executes a file at the given path. Specifically used for `setup.py` files.
    
    Parameters
    ----------
    path : `str`
        Path to the file.
    """
    try:
        module_specification = ModuleSpec('setup', SourceFileLoader('setup', path), origin = path)
        module_specification.has_location = True
        module = module_from_spec(module_specification)
        module_specification.loader.exec_module(module)
    finally:
        try:
            del sys.modules['setup']
        except KeyError:
            pass


def get_project_sources_from_setup(path):
    """
    Gets the project's source(s) from `setup.py` file.
    
    Parameters
    ----------
    path : `str`
        Path to the file.
    
    Returns
    -------
    project_sources : `None | set<str>`
    error_message : `None | str`
    """
    setup_called = False
    module_packages = None
    module_name = None
    
    def setup(*, packages = None, name = None, **keyword_parameters):
        nonlocal setup_called
        nonlocal module_packages
        nonlocal module_name
        
        module_packages = packages
        module_name = name
        setup_called = True
    
    # Mock setuptools
    setup_tools_module = ModuleType('setuptools')
    setup_tools_module.setup = setup
    sys.modules['setuptools'] = setup_tools_module
    
    # Execute setup.py file
    try:
        execute_setup_file(path)
    finally:
        try:
            del sys.modules['setuptools']
        except KeyError:
            pass
    
    if not setup_called:
        return None, '`setup.py` never actually called `setup`.'
        
    return build_project_sources(module_name, module_packages), None


def get_toml_loader():
    """
    Gets toml loader and library name.
    
    Returns
    -------
    loader : `None | FunctionType`
    library_name : `None | str`
    """
    loader = None
    
    # Try toml
    library_name = 'toml'
    try:
        from toml import loads as loader
    except ImportError:
        loader = None
    else:
        return loader, library_name
    finally:
        if (loader is None):
            clean_up_toml_library(library_name)
    
    # Try tomllib
    library_name = 'tomllib'
    try:
        from tomllib import loads as loader
    except ImportError:
        loader = None
    else:
        return loader, library_name
    finally:
        if (loader is None):
            clean_up_toml_library(library_name)
    
    return None, None


def clean_up_toml_library(library_name):
    """
    Cleans up a library. Specifically used for a toml one.
    
    Parameters
    ----------
    library_name : `None | str`
        The name of the library to clean up.
    """
    # Nothing to clean up
    if library_name is None:
        return
    
    library_name_length = len(library_name)
    
    module_names = []
    
    for module_name in sys.modules.keys():
        if (
            module_name.startswith(library_name) and
            (
                (len(module_name) == library_name_length) or
                (module_name[library_name_length] == '.')
            )
        ):
            module_names.append(module_name)
    
    for module_name in module_names:
        try:
            del sys.modules[module_name]
        except KeyError:
            pass


def get_project_sources_from_toml(path):
    """
    Gets the project's source(s) from `pyproject.toml` file.
    
    Parameters
    ----------
    path : `str`
        Path to the file.
    
    Returns
    -------
    project_sources : `None | set<str>`
    error_message : `None | str`
    """
    loader, library_name = get_toml_loader()
    if loader is None:
        return None, 'Failed to read `pyproject.toml`, no `.toml` reader available.'
    
    try:
        with open(path, 'rb') as file:
            loaded = loader(file.read())
    finally:
        loader = None
        clean_up_toml_library(library_name)
    
    
    project_info = loaded.get('project', None)
    if (project_info is None) or (not isinstance(project_info, dict)):
        return None, None
    
    return build_project_sources(project_info.get('name', None), project_info.get('packages', None)), None


def get_target_path_from_parameters(parameters, index):
    """
    Gets target path from the given parameters.
    
    Parameters
    ----------
    parameters : `list<str>`
        System parameters usually.
    index : `int`
        The index to read from.
    
    Returns
    -------
    path_parameter : `str`
        The detected path parameter.
    index : `int`
        The new index to read from next time.
    """
    parameter_count = len(parameters)

    path_parameter = None
    
    if parameter_count > index:
        maybe_path = parameters[index]
        if not maybe_path.startswith('-'):
            path_parameter = maybe_path
            index += 1
    
    return path_parameter, index


def find_top_most_init_directories(path):
    """
    Gets the top-most directory with `__init__.py` or equivalent in it.
    
    Parameters
    ----------
    path : `str`
        The path to originate from.
    
    Returns
    -------
    top_most_init_directories : `None | list<str>`
        The found directory if any.
    """
    directories = []
    
    while True:
        directories.append(path)
        
        parent_directory = get_parent_directory_path(path)
        if parent_directory == path:
            break
        
        path = parent_directory
        continue
    
    chains = []
    chain = None
    
    for directory in directories:
        if not has_init(directory):
            chain = None
            continue
        
        if chain is None:
            chain = []
            chains.append(chain)
        
        chain.append(directory)
    
    if chains:
        return [chain[-1] for chain in chains]


def get_sources_from_project_source_directory(path):
    """
    Gets sources for project source directory.
    
    Parameters
    ----------
    path : `str`
        Path to the directory.
    
    Returns
    -------
    project_sources : `None | set<str>`
    error_message : `None | str`
    """
    setup_file_path = join_paths(path, 'setup.py')
    if is_file(setup_file_path):
        return get_project_sources_from_setup(setup_file_path)
    
    setup_file_path = join_paths(path, 'pyproject.toml')
    if is_file(setup_file_path):
        return get_project_sources_from_toml(setup_file_path)
    
    return None, 'Unknown setup file kind.'


def full_split_path(path):
    """
    Slits the given path removing empty parts.
    
    Parameters
    ----------
    path : `str`
        The path to split.
    
    Returns
    -------
    path_parts : `list<str>`
    """
    return [path_part for path_part in path.split(PATH_SEPARATOR) if path_part]


def get_source_and_target(parameters, index):
    """
    Looks up test source and target.
    
    Parameters
    ----------
    parameters : `list<str>`
        System parameters usually.
    index : `int`
        The index to read from.
    
    Returns
    -------
    source_directory : `str`
    sources : `None | set<str>`
    test_collection_route : `None | list<str>`
    errors : `None | list<(str, str)>`
    index : `int`
    """
    target_path, index = get_target_path_from_parameters(parameters, index)
    
    base_path = get_current_working_directory()
    if target_path is not None:
        base_path = join_paths(base_path, target_path)
    
    top_most_init_directories = find_top_most_init_directories(base_path)
    
    source_directory = None
    sources = None
    errors = None
    
    if top_most_init_directories is not None:
        # Check for setup.py-s
        for top_most_init_directory in top_most_init_directories:
            directory_path = get_parent_directory_path(top_most_init_directory)
            if not is_directory_project_source(directory_path):
                continue
            
            sources, error = get_sources_from_project_source_directory(directory_path)
            if (error is not None):
                if errors is None:
                    errors = []
                
                errors.append((directory_path, error))
            
            if (sources is not None):
                source_directory = directory_path
                break
        
        # Check init-s
        if source_directory is None:
            split = split_path(top_most_init_directories[-1])
            source_directory = split[0]
            sources = {split[1]}
    
    # Check setup.py in current.
    if is_directory_project_source(base_path):
        sources, error = get_sources_from_project_source_directory(base_path)
        if (error is not None):
            if errors is None:
                errors = []
            
            errors.append((base_path, error))
        
        if (sources is not None):
            source_directory = base_path
    
    # check current
    if source_directory is None:
        if is_file(base_path):
            split = split_path(base_path)
            source_directory = split[0]
            sources = {split_extension(split[1])[0]}
        else:
            if errors is None:
                errors = []
            
            errors.append((base_path, 'Could not identify source.'))
    
    # Resolve lookup
    if source_directory is None:
        test_collection_route = None
    else:
        test_collection_route = full_split_path(base_path[len(source_directory):])
    
    return source_directory, sources, test_collection_route, errors, index



__all__ = ('iter_collect_test_files_in',)

from .test_file import TestFile


def iter_collect_test_files_in(entry):
    """
    Iterates over the given file system entry and yields back the collected test files.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    entry : ``FileSystemEntry``
        The entry to iterate over.
    
    Yields
    ------
    test_file : ``TestFile``
    """
    if entry.is_directory():
        yield from iter_collect_test_files_in_directory(entry, is_test_directory_name(entry.get_name()))
        entry.purge()
        return
    
    if is_test_file_name(entry.get_name()):
        yield TestFile(entry)
        return
    
    # no more cases
    return


def iter_collect_test_files_in_directory(entry, within_test_directory):
    """
    Iterates over a directory discovering test files.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    entry : ``FileSystemEntry``
        The entry to iterate over.
    within_test_directory : `bool`
        Defines whether we are in a test directory and whether we should collect test files from it.
    
    Yields
    ------
    test_file : ``TestFile``
    """
    # First check directory
    if within_test_directory:
        directory = None
        
        for sub_entry in entry.iter_entries():
            if sub_entry.is_file():
                name = sub_entry.get_name()
                if name == '__init__.py':
                    directory = TestFile(sub_entry)
                    continue
                
                if is_test_file_name(name):
                    test_file = TestFile(sub_entry)
                    
                    if (directory is None):
                        yield test_file
                        continue
                    
                    directory.feed_sub_file(test_file)
                    continue
        
        if (directory is not None):
            yield directory
    
    else:
        for sub_entry in entry.iter_entries():
            if sub_entry.is_directory():
                yield from iter_collect_test_files_in_directory(sub_entry, is_test_directory_name(sub_entry.get_name()))
                sub_entry.purge()


def is_test_file_name(file_name):
    """
    Returns whether the given file name is the name of a test file.
    
    Parameters
    ----------
    file_name : `str`
        A file's name.
    
    Returns
    -------
    is_test_file_name : `bool`
    """
    if file_name.startswith('_'):
        return False
    
    if file_name == 'test.py':
        return True
    
    if file_name.startswith('test_') and file_name.endswith('.py'):
        return True
    
    if file_name.endswith('_tests.py'):
        return True
    
    return False


def is_test_directory_name(directory_name):
    """
    Returns whether the given directory name is a name of a test directory.
    
    Parameters
    ----------
    directory_name : `str`
        A directory's name.
    
    Returns
    -------
    is_test_directory_name : `bool`
    """
    if directory_name == 'tests':
        return True
    
    if directory_name.startswith(('test_', 'tests_')):
        return True
    
    if directory_name.endswith('_tests'):
        return True
    
    return False



__all__ = ('WrapperCalling',)

import reprlib

from scarletio import copy_docs

from ..contexts import ContextCalling
from ..helpers.hashing import hash_dict, hash_set, hash_tuple, try_hash_method
from ..helpers.un_nesting import un_nest_exceptions

from .wrapper_base import WrapperBase
from .wrapper_conflict import WrapperConflict


MODE_RAISING = 1 << 0
MODE_RETURNING = 1 << 2
MODE_CALL_WITH = 1 << 3


class WrapperCalling(WrapperBase):
    """
    Combined wrapper supporting `call_with`, `returning` and `raising`.
    
    Attributes
    ----------
    wrapped : `object`
        The wrapped test.
    calling_keyword_parameters : `None`, `dict` of (`str`, `object`) items
        Keyword parameters to call the test with if ``.is_call_with``.
    calling_positional_parameters : `None`, `tuple` of `object`
        Positional parameter to call the test with if ``.is_call_with``.
    mode : `int`
        Bitwise flag containing the mode of calling and handling returns.
    raising_accept_subtypes : `bool`
        Whether exception subclasses are accepted as well if ``.is_raising``.
    raising_exceptions : `None`, `set` of ``BaseException``
        The raised exceptions to be expected to be raised if ``.is_raising``.
    raising_where : `None`, `callable`
        Additional check to check the raised exception if ``.is_raising`.
    returning_value : `None`, `object`
        The expected returned value of the test if ``.is_returning``.
    """
    __slots__ = (
        'calling_keyword_parameters', 'calling_positional_parameters', 'mode', 'raising_exceptions',
        'raising_accept_subtypes', 'raising_where', 'returning_value'
    )
    
    def __new__(cls, wrapped = None, *, call_with = None, raising = None, returning = None):
        """
        Creates a new combined test wrapper.
        
        Parameters
        ----------
        wrapped : `None`, `object` = `None`, Optional
            The wrapped test if any.
        call_with : `None`, `tuple` (`tuple` of `object`, `dict` of (`str`, `object`) items) = `None`
                , Optional (Keyword only)
            Whether the test should be called with parameters.
        raising : `None`, `tuple` (`set` of ``BaseException``, `bool`, (`None`, `callable`)) = `None` \
                , Optional (Keyword only)
            Whether the test is raising.
        returning : `None`, `tuple` (`object`) = `None`, Optional (Keyword only)
            Whether the test is returning.
        """
        mode = 0
        
        if raising is None:
            raising_exceptions = None
            raising_accept_subtypes = None
            raising_where = None
        
        else:
            mode |= MODE_RAISING
            raising_exceptions, raising_accept_subtypes, raising_where = raising
        
        
        if returning is None:
            returning_value = None
        
        else:
            mode |= MODE_RETURNING
            returning_value, = returning
        
        
        if call_with is None:
            calling_keyword_parameters = None
            calling_positional_parameters = None
        
        else:
            mode |= MODE_CALL_WITH
            calling_positional_parameters, calling_keyword_parameters = call_with
        
        
        self = object.__new__(cls)
        self.wrapped = wrapped
        
        self.calling_keyword_parameters = calling_keyword_parameters
        self.calling_positional_parameters = calling_positional_parameters
        self.mode = mode
        self.raising_accept_subtypes = raising_accept_subtypes
        self.raising_exceptions = raising_exceptions
        self.raising_where = raising_where
        self.returning_value = returning_value
        
        return self
    
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        for field_added, repr_parts in self._cursed_repr_builder():
            
            mode = self.mode
            repr_parts.append(' (')
            type_field_added = False
            
            if mode & MODE_RAISING:
                if type_field_added:
                    repr_parts.append(', ')
                else:
                    type_field_added = True
                
                repr_parts.append('raising')
            
            if mode & MODE_RETURNING:
                if type_field_added:
                    repr_parts.append(', ')
                else:
                    type_field_added = True
            
                repr_parts.append('returning')
            
            if mode & MODE_CALL_WITH:
                if type_field_added:
                    repr_parts.append(', ')
                else:
                    type_field_added = True
            
                repr_parts.append('call_with')
            
            repr_parts.append(')')
            
            
            if mode & MODE_RAISING:
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                repr_parts.append(' raising_exceptions = ')
                repr_parts.append(repr(self.raising_exceptions))
                
                repr_parts.append(', raising_accept_subtypes = ')
                repr_parts.append(repr(self.raising_accept_subtypes))
                
                raising_where = self.raising_where
                if (raising_where is not None):
                    repr_parts.append(', raising_where = ')
                    repr_parts.append(repr(raising_where))
            
            if mode & MODE_RETURNING:
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                repr_parts.append(', returning_value = ')
                repr_parts.append(reprlib.repr(self.returning_value))
            
            if mode & MODE_CALL_WITH:
                calling_positional_parameters = self.calling_positional_parameters
                if calling_positional_parameters:
                    if field_added:
                        repr_parts.append(',')
                    else:
                        field_added = True
                    
                    repr_parts.append(' calling_positional_parameters = ')
                    repr_parts.append(reprlib.repr(calling_positional_parameters))
                
                calling_keyword_parameters = self.calling_keyword_parameters
                if calling_keyword_parameters:
                    if field_added:
                        repr_parts.append(',')
                    else:
                        field_added = True
                    
                    repr_parts.append(' calling_keyword_parameters = ')
                    repr_parts.append(reprlib.repr(calling_keyword_parameters))
            
        return ''.join(repr_parts)
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        hash_value = WrapperBase.__hash__(self)
        
        if self.is_raising():
            hash_value ^= hash_set(self.raising_exceptions)
            hash_value ^= self.raising_accept_subtypes
            raising_where = self.raising_where
            if (raising_where is not None):
                hash_value ^= try_hash_method(raising_where)
                
        if self.is_returning():
            hash_value ^= try_hash_method(self.returning_value)
        
        if self.is_call_with():
            hash_value ^= hash_tuple(self.calling_positional_parameters)
            hash_value ^= hash_dict(self.calling_keyword_parameters)
        
        return hash_value
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        self_mode = self.mode
        if self_mode != other.mode:
            return False
        
        if self_mode & MODE_RAISING:
            if self.raising_exceptions != other.raising_exceptions:
                return False
            
            if self.raising_accept_subtypes != other.raising_accept_subtypes:
                return False
            
            if self.raising_where != other.raising_where:
                return False
        
        if self_mode & MODE_RETURNING:
            if self.returning_value != other.returning_value:
                return False
        
        if self_mode & MODE_CALL_WITH:
            if self.calling_positional_parameters != other.calling_positional_parameters:
                return False
            
            if self.calling_keyword_parameters != other.calling_keyword_parameters:
                return False
        
        return True
    
    
    @copy_docs(WrapperBase.check_conflicts)
    def check_conflicts(self):
        if self.is_raising() and self.is_returning():
            return WrapperConflict(
                self,
                reason = 'A call wrapper cannot be returning & raising at the same time.',
            )
    
    
    @copy_docs(WrapperBase.check_conflict_with)
    def check_conflict_with(self, other):
        if not isinstance(other, type(self)):
            return
        
        
        self_input_field_count = self.is_call_with()
        self_output_field_count = self.is_raising() + self.is_returning()
        
        other_input_count = other.is_call_with()
        other_output_field_count = other.is_raising() + other.is_returning()
        
        if (self_input_field_count + self_output_field_count) == (other_input_count + other_output_field_count):
            # Both defined input and output, so we accept both.
            return
        
        
        if (self_input_field_count + self_output_field_count) != (other_input_count + other_output_field_count):
            # Both should define input and output at the same time if any does.
            return WrapperConflict(
                self,
                other,
                reason = (
                    'Call wrappers must be either just parameterising / returning / raising, or '
                    'parameterising & (returning / raising) at the same time.'
                )
            )
        
        
        if (self_output_field_count + other_output_field_count) == 2:
            return WrapperConflict(
                self,
                other,
                reason = 'Just returning / raising wrappers cannot be combined.'
            )
    
    
    @copy_docs(WrapperBase.is_ignored_when_testing)
    def is_ignored_when_testing(self):
        return False
    
    
    @copy_docs(WrapperBase.is_mutually_exclusive_with)
    def is_mutually_exclusive_with(self, other):
        if type(self) is not type(other):
            return False
        
        if self.is_call_with() and other.is_call_with():
            return True
        
        return False
    
    
    @copy_docs(WrapperBase.get_context)
    def get_context(self, handle):
        return ContextCalling(handle, self)
    
    
    @property
    def raising_key(self):
        """
        Returns the wrapper's raising key.
        
        Returns
        -------
        raising_key : `None`, `tuple` (`set` of ``BaseException``, `bool`, (`None`, `callable`))
        """
        if self.is_raising():
            return (self.raising_exceptions, self.raising_accept_subtypes, self.raising_where)
    
    
    @property
    def returning_key(self):
        """
        Returns the wrapper's raising key.
        
        Returns
        -------
        returning_key : `None`, `tuple` (`object`)
        """
        if self.is_returning():
            return (self.returning_value, )
    
    
    @property
    def call_with_key(self):
        """
        Returns the wrapper's raising key.
        
        Returns
        -------
        call_with_key : `None`, `tuple` (`tuple` of `object`, `dict` of (`str`, `object`) items)
        """
        if self.is_call_with():
            return (self.calling_positional_parameters, self.calling_keyword_parameters)
    
    
    @classmethod
    def raising_constructor(cls, exception_type, *exception_types, accept_subtypes = True, where = None):
        """
        Creates a new raising wrapper.
        
        Parameters
        ----------
        exception_type : `type<BaseException> | BaseException>
            Exception type to expect.
        *exception_types : `tuple<type<BaseException>, BaseException, ...>`
            Additional exception types.
        accept_subtypes : `bool` = `True`
            Whether subclasses are accepted as well.
        where : `None`, `callable` = `None`, Optional (Keyword only)
            Additional check to check the raised exception.
        
        Raises
        ------
        TypeError
            If an `exception_types`'s type is incorrect.
        ValueError
            If no exception was passed.
        
        Returns
        -------
        self : `instance<self>`
        """
        exception_types = un_nest_exceptions((exception_type, *exception_types))
        if not exception_types:
            raise ValueError(
                'At least 1 exception is required.'
            )
        
        return cls(
            raising = (exception_types, accept_subtypes, where),
        )
    
    
    def raising(self, exception_type, *exception_types, accept_subtypes = True, where = None):
        """
        Creates a new raising wrapper extending self.
        
        Parameters
        ----------
        exception_type : `type<BaseException> | BaseException>
            Exception type to expect.
        *exception_types : `tuple<type<BaseException>, BaseException, ...>`
            Additional exception types.
        accept_subtypes : `bool` = `True`
            Whether subclasses are accepted as well.
        where : `None`, `callable` = `None`, Optional (Keyword only)
            Additional check to check the raised exception.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            If an `exception_types`'s type is incorrect.
        ValueError
            If no exception was passed.
        """
        exception_types = un_nest_exceptions((exception_type, *exception_types))
        if not exception_types:
            raise ValueError(
                'At least 1 exception is required.'
            )
        
        return type(self)(
            self.wrapped,
            raising = (exception_types, accept_subtypes, where),
            returning = self.returning_key,
            call_with = self.call_with_key,
        )
    
    
    @classmethod
    def returning_constructor(cls, returning):
        """
        Creates a new returning wrapper.
        
        Parameters
        ----------
        returning : `object`
            The expected value to be returned.
        
        Returns
        -------
        self : `instance<self>`
        """
        return cls(
            returning = (returning, ),
        )
    
    
    def returning(self, returning_value):
        """
        Creates a new returning wrapper extending self.
        
        Parameters
        ----------
        returning_value : `object`
            The expected value to be returned.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return type(self)(
            self.wrapped,
            raising = self.raising_key,
            returning = (returning_value, ),
            call_with = self.call_with_key,
        )
    
    
    def returning_transformed(self, transformer):
        """
        Creates a new returning wrapper transforming the current parameters of the wrapper.
        
        Parameters
        ----------
        transformer : `callable`
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            The wrapper is not parameterised, or incorrect amount parameters registered.
        """
        if not self.is_call_with():
            raise ValueError(
                f'Call wrapper cannot check for return value if it has no parameters added; self = {self!r}.'
            )
        
        calling_positional_parameters = self.calling_positional_parameters
        calling_keyword_parameters = self.calling_keyword_parameters
        
        if calling_positional_parameters is None:
            if calling_keyword_parameters:
                returning_value = transformer()
            else:
                returning_value = transformer(**calling_keyword_parameters)
        
        else:
            if calling_keyword_parameters:
                returning_value = transformer(*calling_positional_parameters)
            else:
                returning_value = transformer(*calling_positional_parameters, **calling_keyword_parameters)
        
        return self.returning(returning_value)
    
    
    def returning_itself(self):
        """
        Creates a new retuning wrapper checking for whether the input value matches the returned one.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            The wrapper is not parameterised, or incorrect amount parameters registered.
        """
        if not self.is_call_with():
            raise ValueError(
                f'Call wrapper cannot check for return value if it has no parameters added; self = {self!r}.'
            )
        
        calling_positional_parameters = self.calling_positional_parameters
        calling_keyword_parameters = self.calling_keyword_parameters
        
        if len(calling_positional_parameters) + len(calling_keyword_parameters) != 1:
            raise ValueError(
                f'`returning_itself` is only applicable if the wrapper has 1 parameter added; '
                f'self = {self!r}.'
            )
        
        if calling_positional_parameters:
            returning_value = calling_positional_parameters[0]
        else:
            returning_value = next(iter(calling_keyword_parameters.values()))
        
        return type(self)(
            self.wrapped,
            raising = self.raising_key,
            returning = (returning_value, ),
            call_with = self.call_with_key,
        )
    
    
    @classmethod
    def call_with_constructor(cls, *calling_positional_parameters, **calling_keyword_parameters):
        """
        Creates new parameterised wrapper.
        
        Parameters
        ----------
        *calling_positional_parameters : `tuple` of `object`
            Positional parameter to call the test with.
        **calling_keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the test with.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls(
            call_with = (calling_positional_parameters, calling_keyword_parameters),
        )
    
    
    def call_with(self, *calling_positional_parameters, **calling_keyword_parameters):
        """
        Creates a new parameterised wrapper extending self.
        
        Parameters
        ----------
        *calling_positional_parameters : `tuple` of `object`
            Positional parameter to call the test with.
        **calling_keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the test with.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return type(self)(
            self.wrapped,
            raising = self.raising_key,
            returning = self.returning_key,
            call_with = (calling_positional_parameters, calling_keyword_parameters),
        )
    
    
    def is_raising(self):
        """
        Returns whether raised exceptions of the test should be checked.
        
        Returns
        -------
        is_raising : `bool`
        """
        return True if self.mode & MODE_RAISING else False
    
    
    def is_returning(self):
        """
        Returns whether returned value of the test should be checked.
        
        Returns
        -------
        is_returning : `bool`
        """
        return True if self.mode & MODE_RETURNING else False
    
    
    def is_call_with(self):
        """
        Returns whether the test should be called with specific parameters.
        
        Returns
        -------
        is_call_with : `bool`
        """
        return True if self.mode & MODE_CALL_WITH else False



__all__ = ('ContextGarbageCollect',)

from gc import collect as gc_collect

from scarletio import copy_docs

from .base import ContextBase


class ContextGarbageCollect(ContextBase):
    """
    Garbage collection contexts that call `gc.collect`.

    Attributes
    ----------
    wrapper_garbage_collect : ``WrapperGarbageCollect``
        The wrapper that defines when should garbage collection happen.
    """
    __slots__ = ('wrapper_garbage_collect',)
    
    def __new__(cls, wrapper_garbage_collect):
        """
        Creates a new test context.
        
        Parameters
        ----------
        wrapper_garbage_collect : ``WrapperGarbageCollect``
            The wrapper that defines when should garbage collection happen.
        """
        self = object.__new__(cls)
        self.wrapper_garbage_collect = wrapper_garbage_collect
        return self
    
    
    @copy_docs(ContextBase.__repr__)
    def __repr__(self):
        return ''.join(['<', self.__class__.__name__, '>'])
    
    
    @copy_docs(ContextBase.start)
    def start(self):
        if self.wrapper_garbage_collect.before:
            gc_collect()
            gc_collect()
    
    
    @copy_docs(ContextBase.close)
    def close(self, result):
        if self.wrapper_garbage_collect.after:
            gc_collect()
            gc_collect()



__all__ = ('FileRegistrationEvent',)

from .base import FileEventBase
from .constants import IDENTIFIER_FILE_REGISTRATION


class FileRegistrationEvent(FileEventBase):
    """
    Dispatched when a test file is registered.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    file : ``TestFile``
        The respective test file.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_FILE_REGISTRATION



__all__ = ('TestFile', )

from sys import modules
from types import FunctionType

from scarletio import RichAttributeErrorBaseType

from ..environment import apply_environments_for_file_at
from ..helpers.path_repr import get_short_path_repr
from ..test_case import TestCase
from ..wrappers import WrapperBase

from .load_failure import TestFileLoadFailure


def is_test_name(name):
    """
    Returns whether the given name should be collected as a test.
    
    Parameters
    ----------
    name : `str`
        The test's name.
    
    Returns
    -------
    is_test_name : `bool`
    """
    if name == 'test':
        return True
    
    if name.startswith('test_'):
        return True
    
    return False


def is_test(name, value):
    """
    Returns whether the given name - value pair refers to a test.
    
    Parameters
    ----------
    name : `str`
        The name of a a variable.
    value : `object`
        The variable's value.
    
    Returns
    -------
    is_test : `bool`
    """
    if not is_test_name(name):
        return False
    
    if isinstance(value, WrapperBase):
        return value.has_bound_test()
    
    if isinstance(value, FunctionType):
        return True
    
    return False


def _test_case_sort_key(test_case):
    """
    Sort key for test cases.
    
    Parameters
    ----------
    test_case : ``TestKey``
        Test case to get it's sort keys.
    
    Returns
    -------
    sort_key : `str`
    """
    return test_case.name


class TestFile(RichAttributeErrorBaseType):
    """
    Describes a test file.
    
    Attributes
    ----------
    _load_failure : `None`, ``TestFileLoadFailure``
        If loading the test file fails, this attribute is set to details about the occurred exception.
    _module : `None`, `ModuleType`
        The module of the test file. Only set when the first call is made to it.
    _results : `None`, `list` of ``Result``
        Results of already ran tests.
    _sub_files : `None`, `list` of ``TestFile``
        Sub files if we are a directory.
    _test_cases : `None`, `list` of ``TestCase``
        The collected test_cases from the file if any. These test_cases are on collected after calling ``.get_test_cases`` for
        the first time.
    entry : ``FileSystemEntry``
        The test file's respective file's or directory's entry in the file system.
    import_route : `str`
        Import route from the base path to import the file from.
    
    Utility Methods
    ---------------
    - Checks & State
    
        - ``.is_loaded_with_success``
        - ``.is_loaded_with_failure``
        - ``.get_load_failure``
        - ``.is_directory``
        - ``.is_loaded``
    
    - Iterators
    
        - ``.iter_results``
        - ``.iter_passed_results``
        - ``.iter_skipped_results``
        - ``.iter_failed_results``
        - ``.iter_sub_files``
    
    - Counters
    
        - ``.get_test_case_count``
        - ``.get_ran_test_count``
        - ``.get_passed_test_count``
        - ``.get_skipped_test_count``
        - ``.get_failed_test_count``
        - ``.get_test_file_count``
    
    - Internal
        
        - ``.get_module``
        - ``.try_load_test_cases``
        - ``.get_test_cases``
        - ``.iter_test_cases``
        - ``.iter_invoke_test_cases``
        - ``.feed_sub_file``
        - ``.iter_test_files``
    """
    __slots__ = (
        '__weakref__', '_load_failure', '_module', '_results', '_sub_files', '_test_cases', 'entry', 'import_route'
    )
    
    def __new__(cls, entry):
        """
        Creates an new test file.
        
        Parameters
        ----------
        entry : ``FileSystemEntry``
            The test file's respective file's or directory's entry in the file system.
        """
        entry.mark_as_used()
        name = entry.get_name()
        
        path_parts = [*(sub_entry.get_name() for sub_entry in entry.iter_parents())]
        if not name == '__init__.py':
            path_parts.append(name)
            
            if path_parts:
                last_path_part = path_parts[-1]
                if last_path_part.endswith('.py'):
                    path_parts[-1] = last_path_part[:-len('.py')]
        
        
        import_route = '.'.join(path_parts)
        
        self = object.__new__(cls)
        self._load_failure = None
        self._module = None
        self._results = None
        self._sub_files = None
        self._test_cases = None
        self.entry = entry
        self.import_route = import_route
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two test files are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.path == other.path
    
    
    def __hash__(self):
        """Returns the hash value of the test file."""
        return hash(self.path)
    
    
    def __repr__(self):
        """Returns the representation of a test file."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' path = ')
        repr_parts.append(get_short_path_repr(self.path))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    def path(self):
        """
        Returns the full path of the represented file.
        
        Returns
        -------
        path : `str`
        """
        return self.entry.get_path()
    
    
    def get_module(self):
        """
        Gets the module of the test file.
        
        Returns
        -------
        module : `None`, `ModuleType`
        
        Raises
        ------
        TestLoadingError
            Loading test file failed.
        """
        if self._load_failure is not None:
            return None
        
        if self._try_load_module():
            return self._module
        
        return None
    
    
    def _try_load_module(self):
        """
        Tries to load the test file's module.
        
        Called by ``.get_module` when first time getting the module.
        
        Returns
        -------
        loaded : `bool`
            Whether the module loaded.
        """
        import_route = self.import_route
        
        try:
            __import__(import_route)
        except BaseException as err:
            self._load_failure = TestFileLoadFailure(self, err)
            return False
        
        module = modules[import_route]
        self._module = module
        return True
    
    
    def try_load_test_cases(self):
        """
        Loads the file's test_cases. Does method if the test file is a directory.
        
        If loading fails, returns an object representing its failure.
        
        Returns
        -------
        loaded : `bool`
            Whether the tests loaded.
        """
        if self.is_directory():
            return False
        
        module = self.get_module()
        if module is None:
            return False
        
        test_cases = []
        
        for name, value in module.__dict__.items():
            if is_test(name, value):
                test_cases.append(TestCase(self, name, value))
        
        test_cases.sort(key = _test_case_sort_key)
        
        self._test_cases = test_cases
        return True
    
    
    def is_loaded(self):
        """
        Returns whether the test file was loaded.
        
        Returns
        -------
        is_loaded : `bool`
        """
        return self._module is not None
    
    
    def is_loaded_with_success(self):
        """
        Returns whether the tests loaded.
        
        Returns
        -------
        is_loaded_with_success : `bool`
        """
        return self._test_cases is not None
    
    
    def is_loaded_with_failure(self):
        """
        Returns whether loading the test file failed.
        
        Returns
        -------
        is_loaded_with_failure : `bool`
        """
        return self._load_failure is not None
    
    
    def get_load_failure(self):
        """
        Returns the loading failure representing the exception occurred when loading the test file.
        
        Returns
        -------
        load_failure : `None`, ``TestFileLoadFailure``
        """
        return self._load_failure
    
    
    def get_test_cases(self):
        """
        Collects the test cases in the test file.
        
        Returns
        -------
        test_cases : `list` of ``TestCase``
        
        Raises
        ------
        TestLoadingError
            Loading test file failed.
        """
        return [*self.iter_test_cases()]
    
    
    def iter_test_cases(self):
        """
        Iterates over the test_cases of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        test : ``TestCase``
        """
        test_cases = self._test_cases
        if (test_cases is not None):
            yield from test_cases
        
        for sub_file in self.iter_sub_files():
            yield from sub_file.iter_test_cases()
    
    
    def iter_invoke_test_cases(self, environment_manager):
        """
        Iterates over the test cases of the file and invokes them. Yields the test cases' results. If the file is a
        directory will do nothing.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        environment_manager : ``EnvironmentManager``
            Testing environment manager.
        
        Yields
        ------
        result : ``Result``
        """
        if self.is_directory():
            return
        
        results = self._results
        if (results is not None):
            return (yield from results)
        
        environment_manager = apply_environments_for_file_at(environment_manager, self.path)
        for test_case in self.iter_test_cases():
            for result in test_case.iter_invoke(environment_manager):
                if (results is None):
                    results = []
                    self._results = results
                
                results.append(result)
                
                yield result
    
    
    def iter_results(self):
        """
        Iterates over the results of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        results = self._results
        if (results is not None):
            yield from results
        
        for sub_file in self.iter_sub_files():
            yield from sub_file.iter_results()
    
    
    def get_test_case_count(self):
        """
        Returns how much test cases are in the test file collected.
        
        Returns
        -------
        test_case_count : `int`
        """
        test_cases = self._test_cases
        if (test_cases is None):
            test_count = 0
        else:
            test_count = len(test_cases)
        
        for sub_file in self.iter_sub_files():
            test_count += sub_file.get_test_case_count()
        
        return test_count
    
    
    def get_ran_test_count(self):
        """
        Returns how much tests already ran of the file.
        
        Returns
        -------
        ran_test_count : `int`
        """
        results = self._results
        if (results is None):
            ran_test_count = 0
        else:
            ran_test_count = len(results)
        
        for sub_file in self.iter_sub_files():
            ran_test_count += sub_file.get_ran_test_count()
        
        return ran_test_count
    
    
    def iter_passed_results(self):
        """
        Iterates over the passed results of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for result in self.iter_results():
            if result.is_passed():
                yield result
    
    
    def iter_skipped_results(self):
        """
        Iterates over the skipped results of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for result in self.iter_results():
            if result.is_skipped():
                yield result
    
    
    def iter_failed_results(self):
        """
        Iterates over the failed results of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for result in self.iter_results():
            if result.is_failed():
                yield result
    
    
    def iter_informal_results(self):
        """
        Iterates over the informal results of the test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for result in self.iter_results():
            if result.is_informal():
                yield result
    
    
    def get_passed_test_count(self):
        """
        Returns how much test passed.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(result.is_passed() for result in self.iter_results())
    
    
    def get_skipped_test_count(self):
        """
        Returns how much test was skipped.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(result.is_skipped() for result in self.iter_results())
    
    
    def get_failed_test_count(self):
        """
        Returns how much test failed.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(result.is_failed() for result in self.iter_results())
    
    
    def feed_sub_file(self, sub_file):
        """
        Adds a sub-file if the file is a directory.
        
        Parameters
        ----------
        sub_file : ``TestFile``
            The sub file to add.
        
        Returns
        -------
        added : `bool`
            Whether the file was added.
        """
        if not self.is_directory():
            return False
        
        sub_files = self._sub_files
        if (sub_files is None):
            sub_files = []
            self._sub_files = sub_files
        
        sub_files.append(sub_file)
        return True
    
    
    def is_directory(self):
        """
        Returns whether the file is a directory.
        
        Returns
        -------
        is_directory : `bool`
        """
        return self.entry.get_name() == '__init__.py'
    
    
    def iter_sub_files(self):
        """
        Iterates over the sub-files registered under this test file.
        
        This method is an iterable generator.
        
        Yields
        ------
        test_file : ``TestFile``
        """
        sub_files = self._sub_files
        if (sub_files is not None):
            yield from sub_files
    
    
    def get_test_file_count(self):
        """
        Returns how much test files the test file contains including itself.
        
        Returns
        -------
        test_file_count : `int`
        """
        test_file_count = 1
        
        for sub_file in self.iter_sub_files():
            test_file_count += sub_file.get_test_file_count()
        
        return test_file_count
    
    
    def iter_test_files(self):
        """
        Iterates over the test files. This includes self and the optionally registered sub-files.
        
        This method is an iterable generator.
        
        Yields
        ------
        test_file : ``TestFile``
        """
        yield self
        
        for sub_file in self.iter_sub_files():
            yield from sub_file.iter_test_files()



__all__ = ()

from reprlib import repr


def build_result_modifier_parameters(modifier_parameters):
    """
    Builds result modifier parameters. If `modifier_parameters` is `None` the output will be an an empty string.
    
    Parameters
    ----------
    modifier_parameters : `None | (None | list<object>, None | dict<str, object)`
        Positional - keyword parameters pair.
    
    Returns
    -------
    built_value : `str`
    """
    return ''.join(render_result_modifier_parameters([], modifier_parameters))


def render_result_modifier_parameters(into, modifier_parameters):
    """
    Renders the result modifier parameters into the given container.
    
    Parameters
    ----------
    into : `list` of `str`
        String parts to render into.
    modifier_parameters : `None | (None | list<object>, None | dict<str, object)`
        Positional - keyword parameters pair.
    
    Returns
    ----------
    into : `list` of `str`
    """
    if modifier_parameters is None:
        return into
    
    position_parameters, keyword_parameters = modifier_parameters
    if (position_parameters is None) and (keyword_parameters is None):
        return into
    
    into.append('[')
    
    field_added = False
    
    if position_parameters is not None:
        for position_parameter in position_parameters:
            if field_added:
                into.append(', ')
            else:
                field_added = True
            
            into.append(repr(position_parameter))
            continue
    
    if (keyword_parameters is not None):
        for key in sorted(keyword_parameters.keys()):
            value = keyword_parameters[key]
            
            if field_added:
                into.append(', ')
            else:
                field_added = True
            
            into.append(key)
            into.append(' = ')
            into.append(repr(value))
    
    into.append(']')
    
    return into



__all__ = ('ReportBase',)

from scarletio import RichAttributeErrorBaseType


class ReportBase(RichAttributeErrorBaseType):
    """
    Base type for test reports.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new report.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the report's representation."""
        return ''.join(['<', self.__class__.__name__, '>'])
    
    
    def is_failure(self):
        """
        Returns whether the report represents a failure.
        
        Returns
        -------
        is_failure : `bool`
        """
        return False
    
    
    def is_informal(self):
        """
        Returns whether the report is an informal report.
        
        Returns
        -------
        is_informal : `bool`
        """
        return False



__all__ = ('WrapperSkipConditional', )

from scarletio import copy_docs

from .wrapper_skip import WrapperSkip


class WrapperSkipConditional(WrapperSkip):
    """
    Skips the test.

    Attributes
    ----------
    wrapped : `None`, `object`
        The wrapped test.
    skip : `bool`
        Whether the test should be skipped.
    """
    __slots__ = ('skip', )
    
    def __new__(cls, skip):
        """
        Creates a new conditional skip wrapper.
        
        Parameters
        ----------
        skip : `bool`
            Whether the test should be skipped.
        """
        skip = bool(skip)
        
        self = WrapperSkip.__new__(cls)
        self.skip = skip
        return self
    
    
    @copy_docs(WrapperSkip.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} skip={self.skip}>'
    
    
    @copy_docs(WrapperSkip.__eq__)
    def __eq__(self, other):
        if type(self) is type(other):
            if self.wrapped != other.wrapped:
                return False
            
            if self.skip != other.skip:
                return False
            
            return True
        
        if type(other) is WrapperSkip:
            if self.wrapped != other.wrapped:
                return False
            
            if not self.skip:
                return False
            
            return True
        
        return NotImplemented
    
    
    @copy_docs(WrapperSkip.__hash__)
    def __hash__(self):
        return self.skip
    
    
    @copy_docs(WrapperSkip.do_skip)
    def do_skip(self):
        return self.skip



__all__ = ()

from ...result import (
    ReportFailureAsserting, ReportFailureParameterMismatch, ReportFailureRaising, ReportFailureReturning
)

from ..colors import COLOR_FAIL, COLOR_PATH
from ..text_styling import style_text_into

from .result_rendering import (
    render_report_asserting_into, render_report_output_into, render_report_parameter_mismatch_into,
    render_report_raising_into, render_report_returning_into, render_report_reversed_into, render_wrapper_conflict_into
)


def write_load_failure(output_writer, load_failure):
    """
    Writes load failure.
    
    Parameters
    ----------
    output_writer : ``OutputWriter``
        The output writer to write the output with.
    load_failure : ``TestFileLoadFailure``
        Test file load failure.
    """
    message_parts = []
    message_parts = style_text_into(message_parts, 'Exception occurred while loading:\n', COLOR_FAIL)
    message_parts = style_text_into(message_parts, load_failure.path, COLOR_PATH)
    message_parts.append('\n\n')
    message_parts = style_text_into(message_parts, load_failure.exception_message, COLOR_FAIL)
    
    output_writer.write_line(''.join(message_parts))
    output_writer.write_break_line()


def write_result_failing(output_writer, result):
    """
    Writes a failing test.
    
    Parameters
    ----------
    output_writer : ``OutputWriter``
        The output writer to write the output with.
    result : ``Result``
        The failing test.
    """
    message_parts = []
    
    if result.is_conflicted():
        message_parts = render_wrapper_conflict_into(message_parts, result)
    
    elif result.reversed:
        message_parts = render_report_reversed_into(message_parts, result)
    
    else:
        failure_report = result.get_failure_report()
        if isinstance(failure_report, ReportFailureAsserting):
            message_parts = render_report_asserting_into(message_parts, result, failure_report)
        
        elif isinstance(failure_report, ReportFailureRaising):
            message_parts = render_report_raising_into(message_parts, result, failure_report)
        
        elif isinstance(failure_report, ReportFailureReturning):
            message_parts = render_report_returning_into(message_parts, result, failure_report)
        
        elif isinstance(failure_report, ReportFailureParameterMismatch):
            message_parts = render_report_parameter_mismatch_into(message_parts, result, failure_report)
            
        else:
            # No other case.
            return
        
    output_writer.write_line(''.join(message_parts))
    output_writer.write_break_line()


def write_result_informal(output_writer, result):
    """
    Writes the extra information attached to the test.
    
    Parameters
    ----------
    output_writer : ``OutputWriter``
        The output writer to write the output with.
    result : ``Result``
        The informal test.
    """
    output_report = result.get_output_report()
    if output_report is None:
        return
    
    message_parts = []
    render_report_output_into(message_parts, result, output_report)
    output_writer.write_line(''.join(message_parts))
    output_writer.write_break_line()
    



__all__ = ('WrapperChainer',)

from scarletio import copy_docs, export

from ..helpers.hashing import hash_set

from .wrapper_base import WrapperBase


@export
class WrapperChainer(WrapperBase):
    """
    Attributes
    ----------
    wrapped : `None`, `object`
        The wrapped test.
    wrappers : `set` of ``WrapperBase``
        The applied wrappers.
    """
    __slots__ = ('wrappers',)
    
    def __new__(cls, wrapped = None):
        """
        Creates a new combined wrapper containing more wrappers.
        
        Parameters
        ----------
        wrapped : `None`, `object` = `None`, Optional
            The wrapped test if any.
        """
        self = object.__new__(cls)
        self.wrapped = wrapped
        self.wrappers = set()
        return self
    
    
    @copy_docs(WrapperBase.__call__)
    def __call__(self, to_wrap):
        if isinstance(to_wrap, WrapperBase):
            self.append(to_wrap)
            return self
        
        
        if (self.wrapped is not None):
            raise RuntimeError(f'Wrapped already wrapped; self = {self!r}, to_wrap = {to_wrap!r}.')
        
        self.wrapped = to_wrap
        return self
    
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        for field_added, repr_parts in self._cursed_repr_builder():
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' wrappers = ')
            repr_parts.append(repr(self.wrappers))
        
        return ''.join(repr_parts)
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        hash_value = WrapperBase.__hash__(self)
        hash_value ^= hash_set(self.wrappers)
        return hash_value
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.wrappers != other.wrappers:
            return False
        
        return True
    
    
    @copy_docs(WrapperBase.do_skip)
    def do_skip(self):
        for wrapper in self.wrappers:
            if wrapper.do_skip():
                return True
        
        return False
    
    
    @copy_docs(WrapperBase.do_reverse)
    def do_reverse(self):
        for wrapper in self.wrappers:
            if wrapper.do_reverse():
                return True
        
        return False
    
    @copy_docs(WrapperBase.check_conflicts)
    def check_conflicts(self):
        wrappers = list(self.wrappers)
        for wrapper in wrappers:
            wrapper_conflict = wrapper.check_conflicts()
            if (wrapper_conflict is not None):
                return wrapper_conflict
        
        
        while True:
            wrapper_to_check = wrappers.pop()
            
            if not wrappers:
                break
            
            for wrapper in wrappers:
                wrapper_conflict = wrapper_to_check.check_conflict_with(wrapper)
                if (wrapper_conflict is not None):
                    return wrapper_conflict
    
    
    @copy_docs(WrapperBase.iter_wrappers)
    def iter_wrappers(self):
        for wrapper in self.wrappers:
            yield from wrapper.iter_wrappers()
    
    
    @copy_docs(WrapperBase.iter_environments)
    def iter_environments(self):
        for wrapper in self.wrappers:
            yield from wrapper.iter_environments()
    
    
    def append(self, wrapper):
        """
        Adds a new wrapper to chain up.
        
        Parameters
        ----------
        wrapper : ``WrapperBase``
            The wrapper to extend self with.
        """
        if isinstance(wrapper, WrapperChainer):
            for sub_wrapper in wrapper.wrappers:
                self.append(sub_wrapper)
            
            return
        
        
        self.wrappers.add(wrapper)



__all__ = ('get_assertion_state_name', )


ASSERTION_STATE_NONE = 0
ASSERTION_STATE_CREATED = 1
ASSERTION_STATE_FAILED = 2
ASSERTION_STATE_PASSED = 3

DEFAULT_ASSERTION_STATE_NAME = 'unknown'

ASSERTION_STATE_TO_NAME = {
    ASSERTION_STATE_NONE: 'none',
    ASSERTION_STATE_CREATED: 'created',
    ASSERTION_STATE_FAILED: 'failed',
    ASSERTION_STATE_PASSED: 'passed',
}


def get_assertion_state_name(state):
    """
    Returns the condition state's name.
    
    Parameters
    ----------
    state : `int`
        Condition state value.
    
    Returns
    -------
    state_name : `str`
        The condition state's value.
    """
    return ASSERTION_STATE_TO_NAME.get(state, DEFAULT_ASSERTION_STATE_NAME)



__all__ = ('RunState',)

from scarletio import RichAttributeErrorBaseType


class RunState(RichAttributeErrorBaseType):
    """
    Represents a test runner's run state.
    
    Attributes
    ----------
    _load_failures : `None`, `list` of ``TestFileLoadFailure``
        Load failures.
    _loaded_test_files : `None`, `list` of ``TestFile``
        The test files which have been successfully loaded.
    _results : `None`, `list` of ``Result``
        The ran tests' results.
    _test_files : `None`, `list` of ``TestFile``
        The collected test files.
    """
    __slots__ = ('_load_failures', '_loaded_test_files', '_results', '_test_files',)
    
    def __new__(cls):
        """
        Creates a new run state.
        """
        self = object.__new__(cls)
        self._load_failures = None
        self._loaded_test_files = None
        self._results = None
        self._test_files = None
        return self
    
    # generic interaction methods
    
    def get_load_failure_count(self):
        """
        Returns how much much tests files were collected.
        """
        load_failures = self._load_failures
        if (load_failures is None):
            load_failure_count = 0
        else:
            load_failure_count = len(load_failures)
        
        return load_failure_count
    
    
    def iter_load_failures(self):
        """
        Iterates over the load failures.
        
        This method is an iterable generator.
        
        Yields
        ------
        load_failure : ``TestFileLoadFailure``
        """
        load_failures = self._load_failures
        if (load_failures is not None):
            yield from load_failures
    
    
    def get_load_failures(self):
        """
        Returns a copy of the load failures.
        
        Returns
        -------
        load_failures : `list` of ``TestFileLoadFailure``
        """
        load_failures = self._load_failures
        if load_failures is None:
            load_failures = []
        else:
            load_failures = load_failures.copy()
        
        return load_failures
    
    
    def add_load_failure(self, load_failure):
        """
        Adds a new load failure.
        
        Parameters
        ----------
        load_failure : ``TestFileLoadFailure``
        """
        load_failures = self._load_failures
        if load_failures is None:
            load_failures = []
            self._load_failures = load_failures
        
        load_failures.append(load_failure)
    
    

    def get_loaded_test_file_count(self):
        """
        Returns how much much loaded tests files were collected.
        """
        loaded_test_files = self._loaded_test_files
        if (loaded_test_files is None):
            loaded_test_file_count = 0
        else:
            loaded_test_file_count = len(loaded_test_files)
        
        return loaded_test_file_count
    
    
    def iter_loaded_test_files(self):
        """
        Iterates over the loaded  test files.
        
        This method is an iterable generator.
        
        Yields
        ------
        loaded_test_files : ``TestFile``
        """
        loaded_test_files = self._loaded_test_files
        if (loaded_test_files is not None):
            yield from loaded_test_files
    
    
    def get_loaded_test_files(self):
        """
        Returns a copy of the loaded test files.
        
        Returns
        -------
        loaded_test_files : `list` of ``TestFile``
        """
        loaded_test_files = self._loaded_test_files
        if loaded_test_files is None:
            loaded_test_files = []
        else:
            loaded_test_files = loaded_test_files.copy()
        
        return loaded_test_files
    
    
    def add_loaded_test_files(self, loaded_test_file):
        """
        Adds a new loaded test file.
        
        Parameters
        ----------
        loaded_test_file : ``TestFile``
        """
        loaded_test_files = self._loaded_test_files
        if loaded_test_files is None:
            loaded_test_files = []
            self._loaded_test_files = loaded_test_files
        
        loaded_test_files.append(loaded_test_file)
    

    def get_results_count(self):
        """
        Returns how much much tests files were collected.
        
        Returns
        -------
        results_count : `int`
        """
        results = self._results
        if (results is None):
            results_count = 0
        else:
            results_count = len(results)
        
        return results_count
    
    
    def iter_results(self):
        """
        Iterates over the result.
        
        This method is an iterable generator.
        
        Yields
        ------
        results : ``Result``
        """
        results = self._results
        if (results is not None):
            yield from results
    
    
    def get_results(self):
        """
        Returns a copy of the result.
        
        Returns
        -------
        results : `list` of ``Result``
        """
        results = self._results
        if results is None:
            results = []
        else:
            results = results.copy()
        
        return results
    
    
    def add_results(self, result):
        """
        Adds a new result.
        
        Parameters
        ----------
        result : ``Result``
            The result to add.
        """
        results = self._results
        if results is None:
            results = []
            self._results = results
        
        results.append(result)
    
    
    def get_test_file_count(self):
        """
        Returns how much much tests files were collected.
        
        Returns
        -------
        test_file_count : `int`
        """
        test_files = self._test_files
        if (test_files is None):
            test_file_count = 0
        else:
            test_file_count = len(test_files)
        
        return test_file_count
    
    
    def iter_test_files(self):
        """
        Iterates over the test files.
        
        This method is an iterable generator.
        
        Yields
        ------
        test_file : ``TestFile``
        """
        test_files = self._test_files
        if (test_files is not None):
            yield from test_files
    
    
    def get_test_files(self):
        """
        Returns a copy of the test files.
        
        Returns
        -------
        test_files : `list` of ``TestFile``
        """
        test_files = self._test_files
        if test_files is None:
            test_files = []
        else:
            test_files = test_files.copy()
        
        return test_files
    
    
    def add_test_file(self, test_file):
        """
        Adds a new test file.
        
        Parameters
        ----------
        test_file : ``TestFile``
        """
        test_files = self._test_files
        if test_files is None:
            test_files = []
            self._test_files = test_files
        
        test_files.append(test_file)
    
    # Added features
    
    def __repr__(self):
        """Returns the run state's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' loaded_test_files: ')
        repr_parts.append(repr(self.get_loaded_test_file_count()))
        repr_parts.append(' / ')
        repr_parts.append(repr(self.get_test_file_count()))
        
        repr_parts.append(' tests_ran: ')
        repr_parts.append(repr(self.get_results_count()))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_total_registered_test_count(self):
        """
        Returns how much tests were registered by the loaded test files.
        
        Returns
        -------
        total_registered_test_count : `int`
        """
        total_registered_test_count = 0
        
        for test_file in self.iter_loaded_test_files():
            total_registered_test_count += test_file.get_test_count()
        
        return total_registered_test_count
    
    
    def iter_test_cases(self):
        """
        Iterates over the test cases.
        
        Yields
        ------
        test_case : ``TestCase``
        """
        for test_file in self.iter_test_files():
            yield from test_file.iter_tests()



__all__ = ('FileTestingDoneEvent',)

from .base import FileEventBase
from .constants import IDENTIFIER_FILE_TESTING_DONE


class FileTestingDoneEvent(FileEventBase):
    """
    Dispatched when a test file's all tests ran.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    file : ``TestFile``
        The respective test file.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_FILE_TESTING_DONE



__all__ = ('ContextCalling',)

from scarletio import copy_docs, include

from ..helpers.exception_matching import try_match_exception

from .base import ContextBase


AssertionException = include('AssertionException')
Result = include('Result')
ResultState = include('ResultState')


class ContextCalling(ContextBase):
    """
    Base test context.

    Attributes
    ----------
    handle : ``Handle``
        The parent handle.
    wrapper_calling : ``WrapperCalling``
        The wrapper that defines the input and output of the test.
    """
    __slots__ = ('handle', 'wrapper_calling')
    
    def __new__(cls, handle, wrapper_calling):
        """
        Creates a new test context.
        
        Parameters
        ----------
        handle : ``Handle``
            The parent handle.
        wrapper_calling : ``WrapperCalling``
            The wrapper that defines the input and output of the test.
        """
        self = object.__new__(cls)
        self.handle = handle
        self.wrapper_calling = wrapper_calling
        return self
    
    
    @copy_docs(ContextBase.__repr__)
    def __repr__(self):
        return ''.join(['<', self.__class__.__name__, '>'])
    
    
    @copy_docs(ContextBase.enter)
    def enter(self, call_state):
        wrapper_calling = self.wrapper_calling
        if wrapper_calling.is_call_with():
            call_state = call_state.with_parameters(
                wrapper_calling.calling_positional_parameters,
                wrapper_calling.calling_keyword_parameters,
            )
        
        return (None, call_state)
    
    
    @copy_docs(ContextBase.exit)
    def exit(self, result_state):
        wrapper_calling = self.wrapper_calling
        
        if (result_state is None) or (not result_state.is_raise()):
            raised_exception = None
        else:
            raised_exception = result_state.result
        
        result = None
        
        if (raised_exception is None) or (not isinstance(raised_exception, AssertionException)):
            handle = self.handle
            
            if wrapper_calling.is_raising():
                raising_exceptions = wrapper_calling.raising_exceptions
                raising_accept_subtypes = wrapper_calling.raising_accept_subtypes
                
                if raised_exception is None:
                    result = Result(
                        handle.case
                    ).with_handle(
                        handle
                    ).with_exception(
                        raising_exceptions, None, raising_accept_subtypes,
                    )
                
                else:
                    if try_match_exception(
                        raising_exceptions, raised_exception, raising_accept_subtypes, wrapper_calling.raising_where
                    ):
                        if (result_state is not None):
                            result_state = ResultState()
                    
                    else:
                        result = Result(
                            handle.case
                        ).with_handle(
                            handle
                        ).with_exception(
                            raising_exceptions, raised_exception, raising_accept_subtypes,
                        )
            
            elif wrapper_calling.is_returning():
                if (raised_exception is not None):
                    result = Result(handle.case).with_handle(handle).with_exception(None, raised_exception, False)
                
                else:
                    if (result_state is None) or (not result_state.is_return()):
                        returned_value = None
                    else:
                        returned_value = result_state.result
                    
                    returning_value = wrapper_calling.returning_value
                    
                    if returned_value != returning_value:
                        result = Result(handle.case).with_handle(handle).with_return(returning_value, returned_value)
        
        return result, result_state



__all__ = ('WrapperGarbageCollect',)

from scarletio import copy_docs

from ..contexts import ContextGarbageCollect

from .wrapper_base import WrapperBase


class WrapperGarbageCollect(WrapperBase):
    """
    Collects garbage before and (or) after the test as configured.
    
    Attributes
    ----------
    wrapped : `None`, `object`
        The wrapped test.
    after : `bool`
        Whether garbage should be collected after the test.
    before : `bool`
        Whether garbage should be collected before the test.
    """
    __slots__ = ('after', 'before')
    
    def __new__(cls, *, after = False, before = False):
        """
        Creates a new garbage collect wrapper.
        
        Parameters
        ----------
        after : `bool` = `False`, Optional (Keyword only)
            Whether garbage should be collected after the test.
        before : `bool` = `False`, Optional (Keyword only)
            Whether garbage should be collected before the test.
        
        Raises
        ------
        TypeError
            - If `after`'s type is incorrect.
            - If `before`'s type is incorrect.
        """
        if isinstance(after, bool):
            pass
        elif isinstance(after, int):
            after = True if after else False
        else:
            raise TypeError(
                f'`after` can be `bool`, got {after.__class__.__name__}; {after!r}.'
            )
        
        if isinstance(before, bool):
            pass
        elif isinstance(before, int):
            before = True if before else False
        else:
            raise TypeError(
                f'`before` can be `bool`, got {before.__class__.__name__}; {before!r}.'
            )
        
        self = WrapperBase.__new__(cls)
        self.after = after
        self.before = before
        return self
    
    
    @copy_docs(WrapperBase.get_context)
    def get_context(self, handle):
        return ContextGarbageCollect(self)
    
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} before = {self.before!r}, after = {self.after!r}>'
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.after != other.after:
            return False
        
        if self.before != other.before:
            return False
        
        return True
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        return self.after ^ (self.before << 1)



__all__ = ('ContextOutputCapturing',)

import sys
from io import StringIO

from scarletio import copy_docs

from .base import ContextBase


class ContextOutputCapturing(ContextBase):
    """
    Captures output.

    Attributes
    ----------
    standard_error_stream : `None`, `io-like`
        Standard error stream.
    standard_output_stream : `None`, `io-like`
        Standard output stream.
    stream : `None`, ``StringIO``
        Stream to capture stdout and stderr into.
    """
    __slots__ = ('standard_error_stream', 'standard_output_stream', 'stream')
    
    def __new__(cls):
        """
        Creates a new test context.
        """
        self = object.__new__(cls)
        self.stream = None
        self.standard_error_stream = None
        self.standard_output_stream = None
        return self
    
    
    @copy_docs(ContextBase.start)
    def start(self):
        stream = StringIO()
        self.stream = stream
        self.standard_output_stream = sys.stdout
        self.standard_error_stream = sys.stderr
        sys.stdout = stream
        sys.stderr = stream
    
    
    @copy_docs(ContextBase.close)
    def close(self, result):
        stream = self.stream
        if stream is None:
            # Was not started
            return
        
        sys.stdout = self.standard_output_stream
        sys.stderr = self.standard_error_stream
        self.standard_output_stream = None
        self.standard_error_stream = None
        
        if (result is not None):
            output = stream.getvalue()
            if output:
                result.with_output(output)



__all__ = ('TestCase', )

import reprlib

from scarletio import RichAttributeErrorBaseType, WeakReferer

from .handling import Handle
from .helpers.hashing import hash_object
from .result import Result
from .wrappers import WrapperBase


class TestCase(RichAttributeErrorBaseType):
    """
    Represents a test.
    
    Attributes
    ----------
    _test_file_reference : `None`, ``WeakReferer`` to ``TestFile``
        The parent test file of the case.
    name : `str`
        The test's name.
    test : `FunctionType`, ``WrapperBase``
        The test itself, or the wrapped test.
    wrapper : `None`, ``WrapperBase``
        Wrappers containing the test if any.    
    """
    __slots__ = ('_test_file_reference', 'name', 'test', 'wrapper')
    
    def __new__(cls, test_file, name, test):
        """
        Creates a new test case.
        
        Parameters
        ----------
        test_file : ``TestFile``
            The parent test file.
        name : `str`
            The test's name.
        test : `FunctionType`, ``WrapperBase``
            The test itself, or the wrapped test.
        """
        if isinstance(test, WrapperBase):
            wrapper = test
            test = test.unbind_test()
        else:
            wrapper = None
            test = test
        
        self = object.__new__(cls)
        self._test_file_reference = WeakReferer(test_file)
        self.name = name
        self.test = test
        self.wrapper = wrapper
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two test cases are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return False
    
    
    def __hash__(self):
        """Returns the test case's hash value."""
        hash_value = 0
        
        hash_value ^= hash(self.name)
        hash_value ^= hash_object(self.test)
        
        wrapper = self.wrapper
        if (wrapper is not None):
            hash_value ^= hash(wrapper)
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the test case's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', test = ')
        repr_parts.append(reprlib.repr(self.test))
        
        wrapper = self.wrapper
        if (wrapper is not None):
            repr_parts.append(', wrapper = ')
            repr_parts.append(repr(self.wrapper))
        
        if self.do_skip():
            repr_parts.append(', skipped')
        
        if self.do_reverse():
            repr_parts.append(', reversed')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def do_skip(self):
        """
        Returns whether the test should be skipped.
        
        Returns
        -------
        do_skip : `bool`
        """
        wrapper = self.wrapper
        if wrapper is None:
            return False
        
        return wrapper.do_skip()
    
    
    def do_reverse(self):
        """
        Returns whether the test's result should be reversed.
        
        Returns
        -------
        do_reverse : `bool`
        """
        wrapper = self.wrapper
        if wrapper is None:
            return False
        
        return wrapper.do_reverse()
    
    
    def check_conflicts(self):
        """
        Checks the test case's wrappers' conflicts.
        
        Returns
        ------
        wrapper_conflict: `None`, ``WrapperConflict``
        """
        wrapper = self.wrapper
        if (wrapper is not None):
            return wrapper.check_conflicts()
    
    
    def iter_invoke(self, environment_manager):
        """
        Invokes the test case yielding the results of it.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        environment_manager : ``EnvironmentManager``
            Testing environment manager.
        
        Yields
        -------
        result : ``Result``
        """
        conflict = self.check_conflicts()
        if (conflict is not None):
            yield Result(self).with_conflict(conflict)
            return
        
        if self.do_skip():
            yield Result(self).as_skipped()
            return
        
        handles = [*self._iter_handles()]
        for handle, index in zip(handles, reversed(range(len(handles)))):
            result = handle.invoke(environment_manager)
            if index:
                result = result.as_continuous()
            yield result
    
    
    def _iter_handles(self):
        """
        Iterates over the test handles of the test case.
        
        This method is an iterable generator.
        
        Yields
        ------
        handle : ``Handle``
        """
        wrapper = self.wrapper
        if (wrapper is None):
            wrapper_groups = None
            environments = None
        
        else:
            wrapper_groups = set()
            
            environment_wrappers = [*wrapper.iter_environments()]
            if environment_wrappers:
                environments = tuple(environment_wrappers)
            else:
                environments = None
            
            wrappers = [wrapper for wrapper in wrapper.iter_wrappers() if not wrapper.is_ignored_when_testing()]
            
            for initial_wrapper in wrappers:
                wrapper_group = [initial_wrapper]
                
                for wrapper in wrappers:
                    if wrapper in wrapper_group:
                        continue
                    
                    if any(wrapper.is_mutually_exclusive_with(wrapper_to_check) for wrapper_to_check in wrapper_group):
                        continue
                    
                    if any(wrapper_to_check.is_mutually_exclusive_with(wrapper) for wrapper_to_check in wrapper_group):
                        continue
                    
                    wrapper_group.append(wrapper)
                    continue
                
                wrapper_group = frozenset(wrapper_group)
                wrapper_groups.add(wrapper_group)
            
            if not wrapper_groups:
                wrapper_groups = None
        
        test = self.test
        
        if (wrapper_groups is None):
            yield Handle(self, test, None, environments)
        
        else:
            for wrapper_group in wrapper_groups:
                if (wrapper_group is not None):
                    wrapper_group = tuple(wrapper_group)
                
                yield Handle(self, test, wrapper_group, environments)
    
    
    def get_test_file(self):
        """
        Returns the test file of the case.
        
        Returns
        -------
        test_file : `None`, ``TestFile``
        """
        test_file_reference = self._test_file_reference
        if (test_file_reference is not None):
            return test_file_reference()
    
    
    @property
    def import_route(self):
        """
        Returns the import route from the base path to import the file from.
        
        Returns
        -------
        import_route : `str`
        """
        test_file = self.get_test_file()
        if (test_file is None):
            import_route = ''
        else:
            import_route = test_file.import_route
        
        return import_route
    
    
    def is_last(self):
        """
        Returns whether self is the last test case of the file.
        
        Returns
        -------
        is_last : `bool`
        """
        test_file = self.get_test_file()
        if (test_file is None):
            return True
        
        test_cases = test_file._test_cases
        if test_cases is None:
            return True
        
        if self is test_cases[-1]:
            return True
        
        return False



__all__ = ('CaptureOutputContextManager',)

import sys
from io import StringIO

from scarletio import RichAttributeErrorBaseType


class CaptureOutputContextManager(RichAttributeErrorBaseType):
    """
    Captures stdout and stderr.
    
    Attributes
    ----------
    _standard_error_stream : `None`, `io-like`
        Standard error stream.
    _standard_output_stream : `None`, `io-like`
        Standard output stream.
    _stream : `None`, ``StringIO``
        Stream to capture stdout and stderr into.
    """
    __slots__ = ('_standard_error_stream', '_standard_output_stream', '_stream')
    
    def __new__(cls):
        """
        Creates a new output capture instance.
        """
        stream = StringIO()
        self = object.__new__(cls)
        self._standard_error_stream = None
        self._standard_output_stream = None
        self._stream = stream
        return self
    
    
    def __enter__(self):
        """Enters the context manager."""
        if (self._standard_output_stream is not None) or (self._standard_error_stream is not None):
            raise RuntimeError('Context manager already entered.')
        
        self._standard_output_stream = sys.stdout
        self._standard_error_stream = sys.stderr
        stream = self._stream
        sys.stdout = stream
        sys.stderr = stream
        return self
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """Exists the context manager."""
        if (self._standard_output_stream is None) and (self._standard_error_stream is None):
            # Was not started
            return False
        
        sys.stdout = self._standard_output_stream
        sys.stderr = self._standard_error_stream
        self._standard_output_stream = None
        self._standard_error_stream = None
        return False
    
    
    def get_value(self):
        """
        Gets the value written since last call.
        
        Returns
        -------
        value : `str`
        """
        stream = self._stream
        position = stream.tell()
        stream.seek(0)
        value = stream.read(position)
        stream.seek(0)
        return value



__all__ = ('ReportOutput',)

from scarletio import copy_docs

from .base import ReportBase


class ReportOutput(ReportBase):
    """
    Reports output of a test.
    
    Attributes
    ----------
    str : `bool`
    """
    __slots__ = ('output',)
    
    def __new__(cls, output):
        """
        Creates a new report.
        """
        self =  object.__new__(cls)
        self.output = output
        return self
    
    
    def __repr__(self):
        """Returns the report's representation."""
        return ''.join(['<', self.__class__.__name__, ' output = ', repr(self.output), '>'])
    
    
    @copy_docs(ReportBase.is_informal)
    def is_informal(self):
        return True



__all__ = ('TestDoneEvent',)

from .base import ResultEventBase
from .constants import IDENTIFIER_TEST_DONE


class TestDoneEvent(ResultEventBase):
    """
    Dispatched when a test case was executed.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    result : ``Result``
        The respective result.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_TEST_DONE



__all__ = ('ScarletioCoroutineEnvironment',)

from time import sleep as sync_sleep

from scarletio import copy_docs, EventThread

from ..handling import ResultState

from .constants import ENVIRONMENT_TYPE_COROUTINE
from .default import DefaultEnvironment


DEFAULT_TIMEOUT = 60.0


class ScarletioCoroutineEnvironment(DefaultEnvironment):
    """
    Implements a scarletio coroutine environment.
    
    Attributes
    ----------
    timeout : `None`, `int`
        The maximal timeout to interrupt tests before.
        
        Defaults to creating a new event loop every time if set as `None`.
    
    event_loop : `None`, ``EventThread``
        The event loop to use to run the test in.
    
    Class Attributes
    ----------------
    identifier : `int` = `ENVIRONMENT_TYPE_COROUTINE`
        Represents for which environment the test is applicable for.
    """
    __slots__ = ('event_loop', 'timeout',)
    
    identifier = ENVIRONMENT_TYPE_COROUTINE
    
    def __new__(cls, *, event_loop = None, timeout = DEFAULT_TIMEOUT):
        """
        Parameters
        ----------
        event_loop : `None`, ``EventThread`` = `None`, Optional (Keyword only)
            The event loop to use to run the test in.
            S
            Defaults to creating a new event loop every time if given as `None`.
            
        timeout : `None`, `str` = `DEFAULT_TIMEOUT`, Optional (Keyword only)
            The maximal timeout to interrupt tests before.
        """
        self = object.__new__(cls)
        self.event_loop = event_loop
        self.timeout = timeout
        return self
    
    
    @copy_docs(DefaultEnvironment.run)
    def run(self, test, positional_parameters, keyword_parameters):
        event_loop = self.event_loop
        create_event_loop = (event_loop is None)
        
        try:
            if create_event_loop:
                event_loop = EventThread(daemon = True, name = 'scarletio.run', start_later = False)
            
            return event_loop.run(
                self._run_async(test, positional_parameters, keyword_parameters),
                timeout = self.timeout,
            )
        
        finally:
            if create_event_loop:
                if (event_loop is not None):
                    event_loop.stop()
                    event_loop = None
                    sync_sleep(0.0)
    
    
    async def _run_async(self, test, positional_parameters, keyword_parameters):
        """
        Runs the defined test with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        test : `FunctionType`
            The test to call
        positional_parameters : `list` of `object`
            Positional parameters to call the test with.
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the test with.
        
        Returns
        -------
        result_state : ``ResultState``
            The result's product.
        """
        try:
            returned_value = await test(*positional_parameters, **keyword_parameters)
        except BaseException as raised_exception:
            return ResultState().with_raise(raised_exception)
        
        return ResultState().with_return(returned_value)
    
    
    @copy_docs(DefaultEnvironment.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        event_loop = self.event_loop
        if (event_loop is not None):
            repr_parts.append(' event_loop = ')
            repr_parts.append(repr(event_loop))
            
            field_added = True
        
        else:
            field_added = False
        
        timeout = self.timeout
        if (timeout is None) or (timeout != DEFAULT_TIMEOUT):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' timeout = ')
            repr_parts.append(repr(timeout))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(DefaultEnvironment.shutdown)
    def shutdown(self):
        event_loop = self.event_loop
        if (event_loop is not None):
            event_loop.stop()
            sync_sleep(0.0)



__all__ = ()

from scarletio import CallableAnalyzer

from .parameter_mismatch import ParameterMismatch


def iter_parameters_to_pass(positional_parameters, keyword_parameters):
    """
    Iterates over given parameters.
    
    Parameters
    ----------
    positional_parameters : `None | list<object>`
        Positional parameters to call the test with.
    keyword_parameters : `None | dict<str, object> items
        Keyword parameters to call the test with.
    
    This function is an iterable generator.
    
    Yields
    ------
    is_positional : `bool`
        Whether the parameter is a keyword one.
    positional_parameter / keyword_parameter : `object` / `(str, object)`
        They positional or keyword parameter item described by `is_positional`.
    """
    if (positional_parameters is not None):
        for positional_parameter in positional_parameters:
            yield True, positional_parameter
    
    if (keyword_parameters is not None):
        for keyword_parameter in keyword_parameters.items():
            yield False, keyword_parameter


def exhaust_next_positional_parameter(parameters):
    """
    Exhausts and removes (if applicable) the next positional parameter.
    
    Parameters
    ----------
    parameters : `list<Parameter>`
        Parameters to exhaust from.
    
    Returns
    -------
    success : `bool`
    """
    if not parameters:
        return False
    
    parameter = parameters[0]
    if parameter.is_args():
        return True
    
    if parameter.is_positional():
        del parameters[0]
        return True
    
    return False


def exhaust_keyword_parameter(parameters, name):
    """
    Exhausts and removes (if applicable) the a keyword parameter defined by the given name.
    
    Parameters
    ----------
    parameters : `list<Parameter>`
        Parameters to exhaust from.
    name : `str`
        The parameter's name.
    
    Returns
    -------
    success : `bool`
    """
    for index, parameter in enumerate(parameters):
        if parameter.is_keyword():
            if parameter.name == name:
                del parameters[index]
                return True
            
            continue
        
        if parameter.is_kwargs():
            return True
    
    return False


def collect_unsatisfied_parameters(parameters):
    """
    Collects the unsatisfied parameters from the given parameters list.
    
    Parameters
    ----------
    parameters : `list<Parameter>`
        Parameters to collect from.
    
    Returns
    -------
    unsatisfied_parameters : `None | list<Parameter>`
    """
    unsatisfied_parameters = None
    
    for parameter in parameters:
        if parameter.has_default or parameter.is_args() or parameter.is_kwargs():
            continue
        
        if unsatisfied_parameters is None:
            unsatisfied_parameters = []
        
        unsatisfied_parameters.append(parameter)
    
    return unsatisfied_parameters


def check_parameter_mismatch(test, positional_parameters, keyword_parameters):
    """
    Checks whether there is parameter mismatch between the test and the parameters to pass into it.
    
    Parameters
    ----------
    test : `FunctionType`
        The test to check.
    positional_parameters : `None | list<object`
        Positional parameters to call the test with.
    keyword_parameters : `None | dict<str, object>`
        Keyword parameters to call the test with.
    
    Returns
    -------
    parameter_mismatch : `None`, ``ParameterMismatch``
    """
    analyzer = CallableAnalyzer(test)
    original_parameters = analyzer.parameters
    if original_parameters is None:
        original_parameters = []
    else:
        original_parameters = [*original_parameters]
    
    parameters = original_parameters.copy()
    
    
    extra_positional_parameters = None
    extra_keyword_parameters = None
    
    for is_positional, parameter_value in iter_parameters_to_pass(positional_parameters, keyword_parameters):
        if is_positional:
            if exhaust_next_positional_parameter(parameters):
                continue
            
            if extra_positional_parameters is None:
                extra_positional_parameters = []
            
            extra_positional_parameters.append(parameter_value)
        
        else:
            if exhaust_keyword_parameter(parameters, parameter_value[0]):
                continue
            
            if extra_keyword_parameters is None:
                extra_keyword_parameters = {}
            
            extra_keyword_parameters[parameter_value[0]] = parameter_value[1]
    
    
    unsatisfied_parameters = collect_unsatisfied_parameters(parameters)
    
    if (
        (unsatisfied_parameters is None) and
        (extra_positional_parameters is None) and
        (extra_keyword_parameters is None)
    ):
        return None
    
    return ParameterMismatch(
        original_parameters, 
        positional_parameters,
        keyword_parameters,
        unsatisfied_parameters,
        extra_positional_parameters,
        extra_keyword_parameters,
    )



__all__ = ('ResultState',)

from scarletio import RichAttributeErrorBaseType, export

from ..helpers.hashing import hash_object


RESULT_STATE_MODE_NONE = 0
RESULT_STATE_MODE_RETURN = 1
RESULT_STATE_MODE_RAISE = 2


@export
class ResultState(RichAttributeErrorBaseType):
    """
    Represents a test's output.
    
    Attributes
    ----------
    mode : `int`
        The result's mode.
    result : `None`, `object`
        The resulted value.
    """
    __slots__ = ('mode', 'result')
    
    def __new__(cls):
        """
        Creates a new result state.
        """
        self = object.__new__(cls)
        self.mode = RESULT_STATE_MODE_NONE
        self.result = None
        return self
    
    
    def __repr__(self):
        """Returns the representation of the result state."""
        repr_parts = ['<', self.__class__.__name__]
        
        mode = self.mode
        if mode == RESULT_STATE_MODE_NONE:
            field_name = None
        elif mode == RESULT_STATE_MODE_RETURN:
            field_name = 'returned_value'
        elif mode == RESULT_STATE_MODE_RAISE:
            field_name = 'raised_exception'
        else:
            field_name = None
        
        if (field_name is not None):
            repr_parts.append(' ')
            repr_parts.append(field_name)
            repr_parts.append(' = ')
            repr_parts.append(repr(self.result))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two result states are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.mode != other.mode:
            return False
        
        if self.result != other.result:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the result state's hash value."""
        hash_value = 0
        
        hash_value ^= self.mode
        
        result = self.result
        if (result is not None):
            hash_value ^= hash_object(result)
        
        return hash_value
    
    
    def __bool__(self):
        """Returns whether the result state holds anythings."""
        return self.mode != RESULT_STATE_MODE_NONE
    
    
    def copy(self):
        """
        Copies the result state.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.mode = self.mode
        new.result = self.result
        return new
    
    
    def with_return(self, returned_value):
        """
        Creates a new call state overwriting the old one.
        
        Parameters
        ----------
        returned_value : `None`, `object`
            The returned value by the test.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.mode = RESULT_STATE_MODE_RETURN
        new.result = returned_value
        return new
    
    
    def with_raise(self, raised_exception):
        """
        Creates a new call state overwriting the old one.
        
        Parameters
        ----------
        raised_exception : `BaseException`
            The raised exception by the test.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        RuntimeError
            - Raised exception can only be `BaseException` instance.
        """
        if not isinstance(raised_exception, BaseException):
            raise TypeError(
                f'`raised_exception` can only be `{BaseException.__name__}` instance. '
                f'Got {raised_exception.__class__.__name__}; {raised_exception!r}.'
            )
        
        new = object.__new__(type(self))
        new.mode = RESULT_STATE_MODE_RAISE
        new.result = raised_exception
        return new
    
    
    def is_return(self):
        """
        Returns whether the result is a return value.
        
        Returns
        -------
        is_return : `bool`
        """
        return self.mode == RESULT_STATE_MODE_RETURN
    
    
    def is_raise(self):
        """
        Returns whether the result is a raised exception.
        
        Returns
        -------
        is_raise : `bool`
        """
        return self.mode == RESULT_STATE_MODE_RAISE



__all__ = ('FileLoadDoneEvent',)

from .base import FileEventBase
from .constants import IDENTIFIER_FILE_LOAD_DONE


class FileLoadDoneEvent(FileEventBase):
    """
    Dispatched when a test file's loading is done.
    
    Test files can be either be loaded with success or failure. To check it use the
    ``TestFile.is_loaded_with_success`` or the ``TestFile.is_loaded_with_success`` methods.
    
    To get details about why loading the file failed use the ``TestFile.is_loaded_with_success`` method.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    file : ``TestFile``
        The respective test file.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_FILE_LOAD_DONE



__all__ = ('create_default_event_handler_manager',)

from scarletio import RichAttributeErrorBaseType, export

from ..events import (
    FileLoadDoneEvent, FileRegistrationDoneEvent, FileTestingDoneEvent, SourceLoadFailureEvent, TestDoneEvent,
    TestingEndEvent
)
from .base import EventHandlerManager
from .colors import COLOR_FAIL, COLOR_PASS, COLOR_SKIP, COLOR_UNKNOWN
from .default_output_writer import OutputWriter
from .rendering_helpers.load_failure_rendering import render_load_failure_exception
from .rendering_helpers.result_modifier_parameters import build_result_modifier_parameters
from .rendering_helpers.writers import write_load_failure, write_result_failing, write_result_informal
from .text_styling import style_text


@export
def create_default_event_handler_manager():
    """
    Creates a default event handler manager if non is given.
    
    Returns
    ----------
    event_handler_manager : ``EventHandlerManager``
    """
    event_handler_manager = EventHandlerManager()
    
    output_formatter = DefaultEventFormatter()
    event_handler_manager.events(output_formatter.file_registration_done)
    event_handler_manager.events(output_formatter.file_load_done)
    event_handler_manager.events(output_formatter.test_done)
    event_handler_manager.events(output_formatter.file_testing_done)
    event_handler_manager.events(output_formatter.testing_end)
    event_handler_manager.events(output_formatter.source_load_failure)
    
    return event_handler_manager


class DefaultEventFormatter(RichAttributeErrorBaseType):
    """
    Default output formatter for test runner.
    
    Attributes
    ----------
    rendered_entries : `set` of ``FileSystemEntry``
        The  already rendered file system entries.
    output_writer : ``OutputWriter``
        The output writer to write the output with.
    """
    __slots__ = ('rendered_entries', 'output_writer',)
    
    def __new__(cls, output_writer = None):
        """
        Creates a new default event formatter.
        
        Parameters
        ----------
        output_writer : `None`, ``OutputWriter` = `None`, Optional
            The output writer to write the output with.
        """
        if (output_writer is None):
            output_writer = OutputWriter()
        
        self = object.__new__(cls)
        self.rendered_entries = set()
        self.output_writer = output_writer
        return self
    
    
    def __repr__(self):
        """Returns the output formatter representation."""
        return f'<{self.__class__.__name__} output_writer = {self.output_writer!r}>'
    
    
    def file_registration_done(self, event: FileRegistrationDoneEvent):
        """
        Called when all files are registered.
        
        Parameters
        ----------
        event : ``FileRegistrationDoneEvent``
            The dispatched event.
        """
        output_writer = self.output_writer
        output_writer.write_line(f'Collected {event.context.get_registered_file_count()} test file(s).')
        output_writer.write_break_line()
    
    
    def file_load_done(self, event: FileLoadDoneEvent):
        """
        Called when a test file is loaded.
        
        Parameters
        ----------
        event : ``FileLoadDoneEvent``
            The dispatched event.
        """
        file = event.file
        if file.is_loaded_with_failure():
            message_parts = self._maybe_render_test_file_into(
                [],
                file.entry,
                name = style_text(
                    file.entry.get_name(),
                    COLOR_FAIL,
                ),
            )
            
            self.output_writer.write_line(''.join(message_parts))
    
    
    def _maybe_render_test_file_into(self, into, entry, *, name = None):
        """
        Helpers method for rendering file structure tree.
        
        Parameters
        ----------
        into : `list` of `str`
            List to render the structure parts into.
        entry : ``FileSystemEntry``
            The entry to render.
        name : `None`, `str` = `None`, Optional (Keyword only)
            Custom name to use as the entry's name.
        
        Returns
        -------
        into : `list` of `str`
        """
        rendered_entries = self.rendered_entries
        
        if (entry not in rendered_entries):
            for sub_entry in entry.iter_parents():
                if sub_entry in rendered_entries:
                    continue
                    
                into = sub_entry.render_into(into)
                rendered_entries.add(sub_entry)
            
            into = entry.render_into(into, name = name)
            rendered_entries.add(entry)
        
        return into
    
    
    def test_done(self, event: TestDoneEvent):
        """
        Called when a test is done.
        
        Parameters
        ----------
        event : ``TestDoneEvent``
            The dispatched event.
        """
        test_file = event.result.case.get_test_file()
        if (test_file is None):
            # Should not happen
            return
        
        message_parts = self._maybe_render_test_file_into([], test_file.entry)
        
        result = event.result
        if result.is_skipped():
            keyword = 'S'
            foreground = COLOR_SKIP
        
        elif result.is_conflicted():
            keyword = 'C'
            foreground = COLOR_FAIL
        
        elif result.is_informal():
            keyword = 'I'
            foreground = COLOR_PASS
        
        elif result.is_passed():
            keyword = 'P'
            foreground = COLOR_PASS
        
        elif result.is_failed():
            keyword = 'F'
            foreground = COLOR_FAIL
        
        else:
            keyword = '?'
            foreground = COLOR_UNKNOWN
        
        result_modifiers = build_result_modifier_parameters(result.get_modifier_parameters())
        message_parts = test_file.entry.render_custom_sub_directory_into(
            message_parts,
            style_text(f'{keyword} {result.case.name}{result_modifiers}', foreground),
            result.is_last() and result.case.is_last(),
        )
        
        self.output_writer.write_line(''.join(message_parts))
    
    
    def file_testing_done(self, event: FileTestingDoneEvent):
        """
        Called when all tests of a file is done
        
        Parameters
        ----------
        event : ``FileTestingDoneEvent``
            The dispatched event.
        """
        message_parts = self._maybe_render_test_file_into([], event.file.entry)
        if message_parts:
            self.output_writer.write_line(''.join(message_parts))
    
    
    def testing_end(self, event: TestingEndEvent):
        """
        Called when a test is done.
        
        Parameters
        ----------
        event : ``TestDoneEvent``
            The dispatched event.
        """
        output_writer = self.output_writer
        output_writer.write_break_line()
        
        context = event.context
        
        load_failures = context.get_file_load_failures()
        for load_failure in load_failures:
            write_load_failure(output_writer, load_failure)
        
        for result in context.iter_failed_results():
            write_result_failing(output_writer, result)
        
        failed_count = context.get_failed_test_count()
        failed_message_part = f'{failed_count} failed'
        if failed_count:
            failed_message_part = style_text(failed_message_part, COLOR_FAIL)
        else:
            for result in context.iter_informal_results():
                write_result_informal(output_writer, result)
        
        skipped_count = context.get_skipped_test_count()
        skipped_message_part = f'{skipped_count} skipped'
        if skipped_count:
            skipped_message_part = style_text(skipped_message_part, COLOR_SKIP)
        
        
        passed_count = context.get_passed_test_count()
        passed_message_part = f'{passed_count} passed'
        if passed_count:
            passed_message_part = style_text(passed_message_part, COLOR_PASS)
        
        
        output_writer.write(f'{failed_message_part} | {skipped_message_part} | {passed_message_part}')
        if load_failures:
            output_writer.write(
                style_text(f' | {len(load_failures)} files failed to load', COLOR_FAIL))
        output_writer.end_line()
    
    
    def source_load_failure(self, event: SourceLoadFailureEvent):
        """
        Called when a source is failed to load.
        
        Parameters
        ----------
        event : ``TestDoneEvent``
            The dispatched event.
        """
        output_writer = self.output_writer
        output_writer.write_line(f'Failed to import {event.source} from {event.context.runner._source_directory}')
        output_writer.write_break_line()
        output_writer.write(render_load_failure_exception(event.exception))
        output_writer.end_line()



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class ParameterMismatch(RichAttributeErrorBaseType):
    """
    Represents parameter mismatch between a test's parameters and the parameters that would been passed to it.
    
    Attributes
    ----------
    extra_keyword_parameters : `None | dict<str, object>`
        Extra keyword parameters that would been passed.
    extra_positional_parameters : `None | list<object>`
        Extra positional parameters that would been passed.
    keyword_parameters : `None | dict<str, object>`
        Keyword parameters to call the test with.
    parameters : `list<parameter>`
        The test's parameters.
    positional_parameters : `None | list<object>`
        Positional parameters to call the test with.
    unsatisfied_parameters : `None | list<Parameter>`
        Parameters that would nto be satisfied.
    """
    __slots__ = (
        'extra_keyword_parameters', 'extra_positional_parameters', 'keyword_parameters', 'parameters',
        'positional_parameters', 'unsatisfied_parameters'
    )
    
    def __new__(
        cls,
        parameters, 
        positional_parameters,
        keyword_parameters,
        unsatisfied_parameters,
        extra_positional_parameters,
        extra_keyword_parameters,
    ):
        """
        Creates a new parameter mismatch.
        
        Parameters
        ----------
        parameters : `list<parameter>`
            The test's parameters.
        positional_parameters : `None | list<object>`
            Positional parameters to call the test with.
        keyword_parameters : `None | dict<str, object>`
            Keyword parameters to call the test with.
        unsatisfied_parameters : `None | list<Parameter>`
            Parameters that would nto be satisfied.
        extra_positional_parameters : `None | list<object>`
            Extra positional parameters that would been passed.
        extra_keyword_parameters : `None | dict<str, object>`
            Extra keyword parameters that would been passed.
        """
        self = object.__new__(cls)
        self.extra_keyword_parameters = extra_keyword_parameters
        self.extra_positional_parameters = extra_positional_parameters
        self.keyword_parameters = keyword_parameters
        self.parameters = parameters
        self.positional_parameters = positional_parameters
        self.unsatisfied_parameters = unsatisfied_parameters
        return self
    
    
    def __repr__(self):
        """Returns the parameter mismatch's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' parameters = ')
        repr_parts.append(repr(self.parameters))
        
        positional_parameters = self.positional_parameters
        if (positional_parameters is not None):
            repr_parts.append(', positional_parameters = ')
            repr_parts.append(repr(positional_parameters))
        
        keyword_parameters = self.keyword_parameters
        if (keyword_parameters is not None):
            repr_parts.append(', keyword_parameters = ')
            repr_parts.append(repr(keyword_parameters))
        
        unsatisfied_parameters = self.unsatisfied_parameters
        if (unsatisfied_parameters is not None):
            repr_parts.append(', unsatisfied_parameters = ')
            repr_parts.append(repr(unsatisfied_parameters))
            
        extra_positional_parameters = self.extra_positional_parameters
        if (extra_positional_parameters is not None):
            repr_parts.append(', extra_positional_parameters = ')
            repr_parts.append(repr(extra_positional_parameters))
        
        extra_keyword_parameters = self.extra_keyword_parameters
        if (extra_keyword_parameters is not None):
            repr_parts.append(', extra_keyword_parameters = ')
            repr_parts.append(repr(extra_keyword_parameters))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two parameter mismatches are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.extra_keyword_parameters != other.extra_keyword_parameters:
            return False
        
        if self.extra_positional_parameters != other.extra_positional_parameters:
            return False
        
        if self.keyword_parameters != other.keyword_parameters:
            return False
        
        if self.parameters != other.parameters:
            return False
        
        if self.positional_parameters != other.positional_parameters:
            return False
        
        if self.unsatisfied_parameters != other.unsatisfied_parameters:
            return False
        
        return True



__all__ = ('ReportFailureParameterMismatch',)

from scarletio import copy_docs

from .base import ReportBase


class ReportFailureParameterMismatch(ReportBase):
    """
    Test failure representing parameter mismatch.
    
    Attributes
    ----------
    parameter_mismatch : ``ParameterMismatch``
        Parameter mismatch instance.
    """
    __slots__ = ('parameter_mismatch',)
    
    def __new__(cls, parameter_mismatch):
        """
        Creates a new test failure representing a bad return value.
        
        Parameters
        ----------
        parameter_mismatch : ``ParameterMismatch``
            Parameter mismatch instance.
        """
        self = object.__new__(cls)
        self.parameter_mismatch = parameter_mismatch
        return self
    
    
    @copy_docs(ReportBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' parameter_mismatch = ')
        repr_parts.append(repr(self.parameter_mismatch))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ReportBase.is_failure)
    def is_failure(self):
        return True



__all__ = ('WrapperSkip',)

from scarletio import copy_docs

from .wrapper_base import WrapperBase


class WrapperSkip(WrapperBase):
    """
    Skips the test.

    Attributes
    ----------
    wrapped : `None`, `object`
        The wrapped test.
    """
    __slots__ = ()
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__}'
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        return 1
    
    
    @copy_docs(WrapperBase.do_skip)
    def do_skip(self):
        return True



__all__ = ('WrapperCallingFrom',)

import reprlib
from itertools import repeat

from scarletio import copy_docs

from ..helpers.hashing import hash_set, try_hash_method
from ..helpers.un_nesting import un_nest_exceptions

from .wrapper_base import WrapperBase
from .wrapper_call import WrapperCalling
from .wrapper_conflict import WrapperConflict


MODE_RAISING_GIVEN = 1 << 0
MODE_RAISING_LAST = 1 << 2
MODE_RETURNING_GIVEN = 1 << 3
MODE_RETURNING_LAST = 1 << 4
MODE_RETURNING_TRANSFORMED = 1 << 5
MODE_RETURNING_ITSELF = 1 << 6

MODE_RAISING_ALL = MODE_RAISING_GIVEN | MODE_RAISING_LAST
MODE_RETURNING_ALL = MODE_RETURNING_GIVEN | MODE_RETURNING_LAST | MODE_RETURNING_TRANSFORMED | MODE_RETURNING_ITSELF


class WrapperCallingFrom(WrapperBase):
    """
    Allows to create multiple `call_with`, `returning` and `raising` wrappers at once.
    
    Attributes
    ----------
    wrapped : `object`
        The wrapped test.
    calling_from : `list` of `tuple<object>`
        Values to call the tests from.
    mode : `int`
        Bitwise flag containing the mode of handling returns.
    raising_accept_subtypes : `bool`
        Whether exception subclasses are accepted as well if ``.is_raising``.
    raising_exceptions : `None`, `set` of `BaseException`, `list` of `set` of `BaseException`
        The raised exceptions to be expected to be raised if ``.is_raising``.
    raising_where : `None`, `callable`
        Additional check to check the raised exception if ``.is_raising`.
    returning_value : `None`, `object`, `list` of `object`
        The expected returned value of the test if ``.is_returning``.
    """
    __slots__ = (
        'calling_from', 'mode', 'raising_exceptions', 'raising_accept_subtypes', 'raising_where', 'returning_value'
    )
    
    def __new__(cls, wrapped = None, *, calling_from = None, raising = None, returning = None):
        """
        Creates a new combined test wrapper.
        
        Parameters
        ----------
        wrapped : `None`, `object` = `None`, Optional
            The wrapped test if any.
        calling_from : `list` of `tuple<object>`
            Values to call the tests from.
        raising : `None`, `(int, set<BaseException> | list<set<BaseException>>, bool, None | callable)` = `None` \
                , Optional (Keyword only)
            Whether the test is raising.
        returning : `None`, `(int, object | list<object>)` = `None`, Optional (Keyword only)
            Whether the test is returning.
        """
        mode = 0
        
        if raising is None:
            raising_exceptions = None
            raising_accept_subtypes = None
            raising_where = None
        
        else:
            raising_mode, raising_exceptions, raising_accept_subtypes, raising_where = raising
            mode |= raising_mode
        
        if returning is None:
            returning_value = None
        
        else:
            returning_mode, returning_value = returning
            mode |= returning_mode
        
        
        self = object.__new__(cls)
        self.wrapped = wrapped
        self.calling_from = calling_from
        self.mode = mode
        self.raising_accept_subtypes = raising_accept_subtypes
        self.raising_exceptions = raising_exceptions
        self.raising_where = raising_where
        self.returning_value = returning_value
        
        return self
    
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        for field_added, repr_parts in self._cursed_repr_builder():
            mode = self.mode
            
            if mode & MODE_RAISING_ALL:
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                if mode & MODE_RAISING_GIVEN:
                    raising_mode = 'given'
                elif mode & MODE_RAISING_LAST:
                    raising_mode = 'last'
                else:
                    raising_mode = 'unknown'
                
                repr_parts.append(' raising_mode = ')
                repr_parts.append(raising_mode)
                
                repr_parts.append(', raising_exceptions = ')
                repr_parts.append(repr(self.raising_exceptions))
                
                repr_parts.append(', raising_accept_subtypes = ')
                repr_parts.append(repr(self.raising_accept_subtypes))
                
                raising_where = self.raising_where
                if (raising_where is not None):
                    repr_parts.append(', raising_where = ')
                    repr_parts.append(repr(raising_where))
            
            if mode & MODE_RETURNING_ALL:
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                if mode & MODE_RETURNING_GIVEN:
                    returning_mode = 'given'
                elif mode & MODE_RETURNING_LAST:
                    returning_mode = 'last'
                elif mode & MODE_RETURNING_TRANSFORMED:
                    returning_mode = 'transformed'
                elif mode & MODE_RETURNING_ITSELF:
                    returning_mode = 'itself'
                else:
                    returning_mode = 'unknown'
                
                repr_parts.append(' returning_mode = ')
                repr_parts.append(returning_mode)
                
                repr_parts.append(', returning_value = ')
                repr_parts.append(reprlib.repr(self.returning_value))
        
        
        return ''.join(repr_parts)
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        hash_value = WrapperBase.__hash__(self)
        
        hash_value ^= try_hash_method(self.calling_from)
        
        if self.is_raising():
            hash_value ^= hash_set(self.raising_exceptions)
            hash_value ^= self.raising_accept_subtypes
            raising_where = self.raising_where
            if (raising_where is not None):
                hash_value ^= try_hash_method(raising_where)
        
        if self.is_returning():
            returning_value = self.returning_value
            if (returning_value is not None):
                hash_value ^= try_hash_method(returning_value)
        
        return hash_value
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        self_mode = self.mode
        if self_mode != other.mode:
            return False
        
        if self.calling_from != other.calling_from:
            return False
        
        if self_mode & MODE_RAISING_ALL:
            if self.raising_exceptions != other.raising_exceptions:
                return False
            
            if self.raising_accept_subtypes != other.raising_accept_subtypes:
                return False
            
            if self.raising_where != other.raising_where:
                return False
        
        if self_mode & MODE_RETURNING_ALL:
            if self.returning_value != other.returning_value:
                return False
        
        return True
    
    
    @copy_docs(WrapperBase.check_conflicts)
    def check_conflicts(self):
        if self.is_raising() and self.is_returning():
            return WrapperConflict(
                self,
                reason = 'A call-from wrapper cannot be returning & raising at the same time.',
            )
    
    
    # We cant have conflicts
    # @copy_docs(WrapperBase.check_conflict_with)
    # def check_conflict_with(self, other):
    #    pass
    
    
    @copy_docs(WrapperBase.is_ignored_when_testing)
    def is_ignored_when_testing(self):
        return False
    
    
    @copy_docs(WrapperBase.is_mutually_exclusive_with)
    def is_mutually_exclusive_with(self, other):
        if type(self) is type(other):
            return True
        
        if isinstance(other, WrapperCalling):
            if other.is_call_with():
                return True
            
            return False
        
        return False
    
    
    @copy_docs(WrapperBase.iter_wrappers)
    def iter_wrappers(self):
        mode = self.mode
        if mode & MODE_RAISING_ALL:
            if mode & MODE_RAISING_GIVEN:
                raising_key_iterator = repeat(
                    (self.raising_exceptions, self.raising_accept_subtypes, self.raising_where),
                )
            else:
                raising_key_iterator = zip(
                    self.raising_exceptions, repeat(self.raising_accept_subtypes), repeat(self.raising_where)
                )
        else:
            raising_key_iterator = repeat(None)
        
        if mode & MODE_RETURNING_ALL:
            if mode & MODE_RETURNING_GIVEN:
                returning_key_iterator = repeat((self.returning_value,),)
            else:
                returning_key_iterator = zip(self.returning_value)
        else:
            returning_key_iterator = repeat(None)
        
        for calling_with_key, raising_key, returning_key in zip(
            self.calling_from, raising_key_iterator, returning_key_iterator
        ):
            yield WrapperCalling(
                None, call_with = (calling_with_key, {}), raising = raising_key, returning = returning_key
            )
    
    
    @property
    def raising_key(self):
        """
        Returns the wrapper's raising key.
        
        Returns
        -------
        raising_key : `None`, `tuple` (`int`, (`None`, `set` of `BaseException`), `bool`, (`None`, `callable`))
        """
        raising_mode = self.mode & MODE_RAISING_ALL
        if raising_mode:
            return (raising_mode, self.raising_exceptions, self.raising_accept_subtypes, self.raising_where)
    
    
    @property
    def returning_key(self):
        """
        Returns the wrapper's raising key.
        
        Returns
        -------
        returning_key : `None`, `tuple` (`int`, `object`)
        """
        returning_mode = self.mode & MODE_RETURNING_ALL
        if returning_mode:
            return (returning_mode, self.returning_value)
    
    
    
    def raising(self, exception_type, *exception_types, accept_subtypes = True, where = None):
        """
        Creates a new raising wrapper extending self.
        
        Parameters
        ----------
        exception_type : `type<BaseException> | BaseException>
            Exception type to expect.
        *exception_types : `tuple<type<BaseException>, BaseException, ...>`
            Additional exception types.
        accept_subtypes : `bool` = `True`
            Whether subclasses are accepted as well.
        where : `None`, `callable` = `None`, Optional (Keyword only)
            Additional check to check the raised exception.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If an `exception_types`'s type is incorrect.
        ValueError
            - If no exception was passed.
        """
        exception_types = un_nest_exceptions((exception_type, *exception_types))
        if not exception_types:
            raise ValueError(
                'At least 1 exception is required.'
            )
        
        return type(self)(
            self.wrapped,
            calling_from = self.calling_from,
            raising = (MODE_RAISING_GIVEN, exception_types, accept_subtypes, where),
            returning = self.returning_key,
        )
    
    
    def raising_last(self, *, accept_subtypes = True, where = None):
        """
        Creates a new raising last wrapper extending self.
        
        Note that this also removes the last parameter when calling.
        
        Parameters
        ----------
        accept_subtypes : `bool` = `True`
            Whether subclasses are accepted as well.
        where : `None`, `callable` = `None`, Optional (Keyword only)
            Additional check to check the raised exception.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If an `exception_types`'s type is incorrect.
        ValueError
            - If an item has no parameters.
        """
        calling_from = self.calling_from
        if any(len(item) < 1 for item in calling_from):
            raise ValueError(
                f'`raising_last` is only applicable if the wrapper has at least 1 for each call; '
                f'self = {self!r}.'
            )
        
        raises = [un_nest_exceptions(item[-1]) for item in calling_from]
        if any((not raises) for raises in raises):
            raise ValueError(
                'At least 1 exception is required for each case.'
            )
        
        calling_from = [item[:-1] for item in calling_from]
        
        return type(self)(
            self.wrapped,
            calling_from = calling_from,
            raising = (MODE_RAISING_LAST, raises, accept_subtypes, where),
            returning = self.returning_key,
        )
    
    
    def returning(self, returning_value):
        """
        Creates a new returning wrapper extending self.
        
        Parameters
        ----------
        returning_value : `object`
            The expected value to be returned.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return type(self)(
            self.wrapped,
            calling_from = self.calling_from,
            raising = self.raising_key,
            returning = (MODE_RETURNING_GIVEN, returning_value),
        )
    
    
    def returning_last(self):
        """
        Creates a new returning last wrapper extending self.
        
        Note that this also removes the last parameter when calling.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            - If an item has no parameters.
        """
        calling_from = self.calling_from
        if any(len(item) < 1 for item in calling_from):
            raise ValueError(
                f'`returning_last` is only applicable if the wrapper has at least 1 for each call; '
                f'self = {self!r}.'
            )
        
        returns = [item[-1] for item in calling_from]
        calling_from = [item[:-1] for item in calling_from]
        
        return type(self)(
            self.wrapped,
            calling_from = calling_from,
            raising = self.raising_key,
            returning = (MODE_RETURNING_LAST, returns),
        )
    
    
    def returning_transformed(self, transformer):
        """
        Creates a new returning wrapper transforming the current parameters of the wrapper.
        
        Parameters
        ----------
        transformer : `callable`
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        calling_from = self.calling_from
        
        returns = [transformer(*item) for item in calling_from]
        
        return type(self)(
            self.wrapped,
            calling_from = self.calling_from,
            raising = self.raising_key,
            returning = (MODE_RETURNING_TRANSFORMED, returns),
        )
    
    
    def returning_itself(self):
        """
        Creates a new retuning wrapper checking for whether the input value matches the returned one.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        ValueError
            - If an item stores incorrect amount parameters.
        """
        calling_from = self.calling_from
        if any(len(item) != 1 for item in calling_from):
            raise ValueError(
                f'`returning_itself` is only applicable if the wrapper has 1 parameter added for each call; '
                f'self = {self!r}.'
            )
        
        returns = [item[0] for item in calling_from]
        
        return type(self)(
            self.wrapped,
            calling_from = calling_from,
            raising = self.raising_key,
            returning = (MODE_RETURNING_ITSELF, returns),
        )
    
    
    @classmethod
    def calling_from_constructor(cls, calling_from):
        """
        Creates new parameterised wrapper.
        
        Parameters
        ----------
        calling_from : `iterable`
            Iterable to call the test from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if not hasattr(calling_from, '__iter__'):
            raise TypeError(
                f'`calling_from` must be iterable, got {calling_from.__class__.__name__}; {calling_from!r}.'
            )
        
        items = []
        
        for item in calling_from:
            if isinstance(item, tuple):
                items.append(item)
            else:
                items.append((item,),)
        
        
        return cls(
            calling_from = items,
        )
    
    
    def is_raising(self):
        """
        Returns whether raised exceptions of the test should be checked.
        
        Returns
        -------
        is_raising : `bool`
        """
        return True if self.mode & MODE_RAISING_ALL else False
    
    
    def is_returning(self):
        """
        Returns whether returned value of the test should be checked.
        
        Returns
        -------
        is_returning : `bool`
        """
        return True if self.mode & MODE_RETURNING_ALL else False



__all__ = ('RunnerContext',)

from scarletio import RichAttributeErrorBaseType


class RunnerContext(RichAttributeErrorBaseType):
    """
    Represents a test's context. Test contexts are present while running a test.
    
    Attributes
    ----------
    _registered_files : `None`, `list` of ``TestFile``
        The collected test files.
    file_system_entries : `list<FileSystemEntry>`
        The file system entries built with the test runner's settings.
    runner : ``TestRunner``
        The respective test runner running tests.
    
    Utility Methods
    ---------------
    - Count getters
    
        - ``.get_registered_file_count``
        - ``.get_test_case_count``
        - ``.get_ran_test_count``
        - ``.get_passed_test_count``
        - ``.get_skipped_test_count``
        - ``.get_failed_test_count``
        - ``.get_load_failed_file_count``
        - ``.get_load_succeeded_file_count``
    
    - Iterators
    
        - ``.iter_registered_files``
        - ``.iter_results``
        - ``.iter_passed_results``
        - ``.iter_skipped_results``
        - ``.iter_failed_results``
        - ``.iter_load_failed_files``
        - ``.iter_load_succeeded_files``
        - ``.iter_file_load_failures``
    
    - Getters
    
        - ``.get_registered_files``
        - ``.get_failed_to_load``
        - ``.get_file_load_failures``
        - ``.get_skipped_results``
        - ``.get_failed_results``
        - ``.get_load_failed_file_count``
    
    - Internal
    
        - ``.register_file``
    """
    __slots__ = ('_registered_files', 'file_system_entries', 'runner',)
    
    def __new__(cls, runner, file_system_entries):
        """
        Creates a new runner context.
        
        Parameters
        ----------
        runner : ``TestRunner``
            The respective test runner running tests.
        file_system_entries : `list<FileSystemEntry>`
            The file system entries built with the test runner's settings.
        """
        self = object.__new__(cls)
        self._registered_files = None
        self.file_system_entries = file_system_entries
        self.runner = runner
        return self
    
    
    def get_registered_file_count(self):
        """
        Returns how much much registered files were collected.
        """
        registered_file_count = 0
        
        for registered_file in self.iter_registered_files():
            registered_file_count += registered_file.get_test_file_count()
        
        return registered_file_count
    
    
    def iter_registered_files_shallow(self):
        """
        Shallow iterates over the registered files.
        
        This method is an iterable generator.
        
        Yields
        ------
        registered_file : ``TestFile``
        """
        registered_files = self._registered_files
        if (registered_files is not None):
            yield from registered_files
    
    
    def iter_registered_files(self):
        """
        Iterates over the registered files.
        
        This method is an iterable generator.
        
        Yields
        ------
        registered_file : ``TestFile``
        """
        registered_files = self._registered_files
        if (registered_files is not None):
            for registered_file in registered_files:
                yield from registered_file.iter_test_files()
    
    
    def get_registered_files(self):
        """
        Returns a copy of the registered files.
        
        Returns
        -------
        registered_files : `list` of ``TestFile``
        """
        return [*self.iter_registered_files()]
    
    
    def register_file(self, registered_file):
        """
        Adds a new registered file.
        
        Parameters
        ----------
        registered_file : ``TestFile``
        """
        registered_files = self._registered_files
        if registered_files is None:
            registered_files = []
            self._registered_files = registered_files
        
        registered_files.append(registered_file)
    
    
    def get_test_case_count(self):
        """
        Returns how much test cases are in the test files collected.
        
        Returns
        -------
        test_case_count : `int`
        """
        return sum(test_file.get_test_case_count() for test_file in self.iter_registered_files())
    
    
    def get_ran_test_count(self):
        """
        Returns how much tests already ran.
        
        Returns
        -------
        ran_test_count : `int`
        """
        return sum(test_file.get_ran_test_count() for test_file in self.iter_registered_files())
    
    
    def iter_results(self):
        """
        Iterates over the results.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for test_file in self.iter_registered_files():
            yield from test_file.iter_results()
    
    
    def iter_passed_results(self):
        """
        Iterates over the passed results of the runner context.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for test_file in self.iter_registered_files():
            if not test_file.is_directory():
                yield from test_file.iter_passed_results()
    
    
    def iter_skipped_results(self):
        """
        Iterates over the skipped results of the runner context.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for test_file in self.iter_registered_files():
            if not test_file.is_directory():
                yield from test_file.iter_skipped_results()
    
    
    def iter_failed_results(self):
        """
        Iterates over the failed results of the runner context.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for test_file in self.iter_registered_files():
            if not test_file.is_directory():
                yield from test_file.iter_failed_results()
    
    
    def iter_informal_results(self):
        """
        Iterates over the informal only results of the runner context.
        
        This method is an iterable generator.
        
        Yields
        ------
        result : ``Result``
        """
        for test_file in self.iter_registered_files():
            if not test_file.is_directory():
                yield from test_file.iter_informal_results()
    
    
    def get_passed_test_count(self):
        """
        Returns how much test passed.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(
            test_file.get_passed_test_count()
            for test_file in self.iter_registered_files()
            if not test_file.is_directory()
        )
    
    
    def get_skipped_test_count(self):
        """
        Returns how much test was skipped.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(
            test_file.get_skipped_test_count()
            for test_file in self.iter_registered_files()
            if not test_file.is_directory()
        )
    
    
    def get_failed_test_count(self):
        """
        Returns how much test failed.
        
        Returns
        -------
        passed_test_count : `int`
        """
        return sum(
            test_file.get_failed_test_count()
            for test_file in self.iter_registered_files()
            if not test_file.is_directory()
        )
    
    
    def get_passed_results(self):
        """
        Iterates over the passed results of the runner context.
        
        Returns
        -------
        results : `list` of ``Result``
        """
        return [*self.iter_passed_results()]
    
    
    def get_skipped_results(self):
        """
        Iterates over the skipped results of the runner context.
        
        Returns
        -------
        results : `list` of ``Result``
        """
        return [*self.iter_skipped_results()]
    
    
    def get_failed_results(self):
        """
        Iterates over the failed results of the runner context.
        
        Returns
        -------
        results : `list` of ``Result``
        """
        return [*self.iter_failed_results()]
    
    
    def get_load_failed_file_count(self):
        """
        Returns how much files failed to load.
        
        Returns
        -------
        load_failed_file_count : `bool`
        """
        return sum(test_file.is_loaded_with_failure() for test_file in self.iter_registered_files())
    
    
    def get_load_succeeded_file_count(self):
        """
        Returns how much files failed to load.
        
        Returns
        -------
        load_succeeded_file_count : `bool`
        """
        return sum(test_file.is_loaded_with_success() for test_file in self.iter_registered_files())
    
    
    def get_not_loaded_file_count(self):
        """
        Returns how much file was not loaded.
        
        Returns
        -------
        not_loaded_file_count : `bool`
        """
        return sum((not test_file.is_loaded()) for test_file in self.iter_registered_files())
    
    
    def iter_load_failed_files(self):
        """
        Iterates over the files which failed to load.
        
        This method is an iterable generator.
        
        Yields
        -------
        test_file : ``TestFile``
        """
        for test_file in self.iter_registered_files():
            if test_file.is_loaded_with_failure():
                yield test_file
    
    
    def iter_load_succeeded_files(self):
        """
        Iterates over the files which succeeded to load.
        
        This method is an iterable generator.
        
        Yields
        -------
        test_file : ``TestFile``
        """
        for test_file in self.iter_registered_files():
            if test_file.is_loaded_with_success():
                yield test_file
    
    
    def iter_file_load_failures(self):
        """
        Iterates over the file load failures.
        
        This method is an iterable generator.
        
        Yields
        -------
        load_failure : ``TestFileLoadFailure``
        """
        for test_file in self.iter_registered_files():
            load_failure = test_file.get_load_failure()
            if (load_failure is not None):
                yield load_failure
    
    
    def get_file_load_failures(self):
        """
        Returns the file load failures.
        
        Returns
        -------
        load_failures : `list` of ``TestFileLoadFailure``
        """
        return [*self.iter_file_load_failures()]



__all__ = ('FileSystemEntry',)

from os import listdir as list_directory
from os.path import join as join_paths, isdir as is_directory, isfile as is_file

from scarletio import RichAttributeErrorBaseType, WeakReferer


PYTHON_EXTENSIONS = ('.py', '.pyd', '.pyc', '.so')

def is_python_file(path):
    """
    Returns whether the given path refers to a python file.
    
    Parameters
    ----------
    path : `str`
        Path to the file.
    
    Returns
    -------
    path : `str`
    """
    if is_file(path):
        return path
    
    if not path.endswith(PYTHON_EXTENSIONS):
        for extension in PYTHON_EXTENSIONS:
            file_path = path + extension
            if is_file(file_path):
                return file_path
    
    return None


class FileSystemEntry(RichAttributeErrorBaseType):
    __slots__ = (
        '__weakref__', '_directory', '_directory_path', '_entries', '_full_path', '_name', '_parent_reference',
        '_self_reference', '_used'
    )
        
    def __new__(cls, path, name, limit_lookup_to):
        """
        Creates a new File system entry.
        
        Parameters
        ----------
        path : `str`
            Path to the entry's directory.
        name : `str | None`
            The name of the entry.
        limit_lookup_to : `None | list<str>`
            Limits sub-directory lookups to only the given path.
        
        Returns
        -------
        entry : `None | instance<cls>`
        """
        full_path = join_paths(path, name)
        directory = is_directory(full_path)
        
        entries = None
        
        if directory:
            if (limit_lookup_to is None) or (not limit_lookup_to):
                for sub_name in sorted(list_directory(full_path)):
                    entry = cls(full_path, sub_name, None)
                    if entry is None:
                        continue
                    
                    if entries is None:
                        entries = []
                    
                    entries.append(entry)
            
            else:
                entry = cls(full_path, limit_lookup_to[0], limit_lookup_to[1:])
                if (entry is not None):
                    if entries is None:
                        entries = []
                    
                    entries.append(entry)
        else:
            file_path = is_python_file(full_path)
            if file_path is None:
                return None
            
            full_path = file_path
        
        self = object.__new__(cls)
        self._directory = directory
        self._directory_path = path
        self._entries = entries
        self._full_path = full_path
        self._name = name
        self._parent_reference = None
        self._self_reference = None
        self._used = 0
        
        for entry in self.iter_entries():
            entry.link_parent(self)
        
        return self
    
    
    def __bool__(self):
        """Returns whether the entry is used anywhere."""
        if self._used:
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the entry's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' path = ')
        repr_parts.append(repr(self.get_path()))
        
        repr_parts.append(', entry count: ')
        entries = self._entries
        if (entries is None):
            entry_count = 0
        else:
            entry_count = len(entries)
        repr_parts.append(repr(entry_count))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def mark_as_used(self):
        """
        Marks the file as used.
        """
        self._used += 1
        
        parent = self.get_parent()
        if (parent is not None):
            parent.mark_as_used()
    
    
    def get_path(self):
        """
        Returns the full path of the entry.
        
        Returns
        -------
        path : `str`
        """
        return self._full_path
    
    
    def get_self_reference(self):
        """
        Returns the a weak reference to self.
        
        Returns
        -------
        weak_reference : ``WeakReferer``
        """
        self_reference = self._self_reference
        if (self_reference is None):
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
        
        return self_reference
    
    
    def link_parent(self, parent):
        """
        Links the parent directory to self.
        
        Parameters
        ----------
        parent : ``FileSystemEntry``
            The parent to link.
        """
        self._parent_reference = parent.get_self_reference()
    
    
    def get_parent(self):
        """
        Returns the entry's parent.
        
        Returns
        -------
        parent : `None`, ``FileSystemEntry``
        """
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            return parent_reference()
    
    
    def is_last(self):
        """
        Returns whether the entry is last in it's parent.
        
        Returns
        -------
        is_last : `bool`
        """
        parent = self.get_parent()
        if parent is None:
            return True
        
        entries = parent._entries
        if (entries is None):
            return True
        
        if self is entries[-1]:
            return True
        
        return False
    
    
    def iter_parents(self):
        """
        Iterates over the parents of the entry.
        
        This method is an iterable generator.
        
        Yields
        ------
        parent : ``FileSystemEntry``
        """
        parent = self.get_parent()
        if (parent is not None):
            yield from parent.iter_parents()
            yield parent
    
    
    def iter_parents_skip_first(self):
        """
        Iterates over the parents of the entry. But skips the first!
        
        This method is an iterable generator.
        
        Yields
        ------
        parent : ``FileSystemEntry``
        
        Returns
        -------
        reached_end : `bool`
            Whether the parent reached the end.
        """
        parent = self.get_parent()
        if (parent is None):
            return True
        
        reached_end = yield from parent.iter_parents_skip_first()
        if (not reached_end):
            yield parent
        
        return False
    
    
    def has_parents(self):
        """
        Returns whether self has parents.
        
        Returns
        -------
        has_parents : `bool`
        """
        return self.get_parent() is not None
    
    
    def get_name(self):
        """
        Returns the entry's name.
        
        Returns
        -------
        name : `str`
        """
        return self._name
    
    
    def is_directory(self):
        """
        Returns whether self represents a directory.
        
        Returns
        -------
        is_directory : `bool`
        """
        return self._directory
    
    
    def is_file(self):
        """
        Returns whether self represents a file.
        
        Returns
        -------
        is_file : `bool`
        """
        return not self._directory
    
    
    def iter_entries(self):
        """
        Iterates over the sub entries.
        
        This method is an iterable generator.
        
        Yields
        ------
        parent : ``FileSystemEntry``
        """
        entries = self._entries
        if (entries is not None):
            yield from entries
    
    
    def purge(self):
        """
        Prunes all the sub members who are not used.
        
        Returns
        -------
        used : `bool`
        """
        if not self:
            return False
        
        entries = None
        
        for entry in self.iter_entries():
            if entry.purge():
                if entries is None:
                    entries = []
                
                entries.append(entry)
        
        self._entries = entries
        
        return True
    
    
    def render_into(self, into, *, name = None):
        """
        Renders the path access into the given list of strings.
        
        Parameters
        ----------
        into : `list` of `str`
            List to render self into.
        name : `None`, `str` = `None`, Optional (Keyword only)
            Custom name to use.
        
        Returns
        -------
        into : `list` of `str`
        """
        for parent in self.iter_parents_skip_first():
            if parent.is_last():
                into.append('   ')
            else:
                into.append('│  ')
        
        if self.has_parents():
            if self.is_last():
                into.append('└─ ')
            else:
                into.append('├─ ')
        
        if name is None:
            name = self._name
        
        into.append(name)
        into.append('\n')
        
        return into
    
    
    def render_custom_sub_directory_into(self, into, name, is_last):
        """
        Renders a custom sub directory.
        
        Parameters
        ----------
        into : `list` of `str`
            List to render self into.
        name : `str`
            The name of the sub-directory.
        is_last : `bool`
            Whether self is the last element on the line.
        
        Returns
        -------
        into : `list` of `str`
        """
        for parent in self.iter_parents_skip_first():
            if parent.is_last():
                into.append('   ')
            else:
                into.append('│  ')
        
        if self.has_parents():
            if self.is_last():
                into.append('   ')
            else:
                into.append('│  ')
        
        if is_last:
            into.append('└─ ')
        else:
            into.append('├─ ')
        
        into.append(name)
        into.append('\n')
        
        return into



__all__ = ('CallState',)

from scarletio import RichAttributeErrorBaseType

from ..helpers.hashing import hash_dict, hash_list
from ..helpers.merging import maybe_merge_iterables, maybe_merge_mappings


class CallState(RichAttributeErrorBaseType):
    """
    Defines how the function should be called.
    
    Attributes
    ----------
    keyword_parameters : `None`, `dict` of (`str`, `object`) items
        Keyword parameters to the call the test function with.
    positional_parameters : `None`, `list` of `object`
        Positional parameters to the the test function with.
    """
    __slots__ = ('keyword_parameters', 'positional_parameters')
    
    def __new__(cls):
        """
        Creates a new calls state.
        """
        self = object.__new__(cls)
        self.keyword_parameters = None
        self.positional_parameters = None
        return self
    
    
    def __repr__(self):
        """Returns the representation of the call state."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        positional_parameters = self.positional_parameters
        if (positional_parameters is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' positional_parameters = ')
            repr_parts.append(repr(positional_parameters))
        
        keyword_parameters = self.keyword_parameters
        if (keyword_parameters is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' keyword_parameters = ')
            repr_parts.append(repr(keyword_parameters))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two call states are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.keyword_parameters != other.keyword_parameters:
            return False
        
        if self.positional_parameters != other.positional_parameters:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the call state's hash value."""
        hash_value = 0
        
        keyword_parameters = self.keyword_parameters
        if (keyword_parameters is not None):
            hash_value ^= 1 << 4
            hash_value ^= hash_dict(keyword_parameters)
        
        positional_parameters = self.positional_parameters
        if (positional_parameters is not None):
            hash_value ^= 1 << 8
            hash_value ^= hash_list(positional_parameters)
        
        return hash_value
    
    
    def __bool__(self):
        """Returns whether the call state holds anythings."""
        if (self.keyword_parameters is not None):
            return True
        
        if (self.positional_parameters is not None):
            return True
        
        return False
    
    
    def copy(self):
        """
        Copies the call state.
        
        Returns
        -------
        new : `instance<type<self>`
        """
        keyword_parameters = self.keyword_parameters
        if (keyword_parameters is not None):
            keyword_parameters = keyword_parameters.copy()
        
        positional_parameters = self.positional_parameters
        if (positional_parameters is not None):
            positional_parameters = positional_parameters.copy()
        
        new = object.__new__(type(self))
        new.keyword_parameters = keyword_parameters
        new.positional_parameters = positional_parameters
        return new
    
    
    def with_parameters(self, positional_parameters, keyword_parameters):
        """
        Creates a new call state with merged parameters.
        
        Parameters
        ----------
        positional_parameters : `None`, `list` of `object`
            Positional parameters to the the test function with.
        keyword_parameters : `None`, `dict` of (`str`, `object`) items
            Keyword parameters to the call the test function with.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        positional_parameters = maybe_merge_iterables(self.positional_parameters, positional_parameters)
        keyword_parameters = maybe_merge_mappings(self.keyword_parameters, keyword_parameters)
        
        new = object.__new__(type(self))
        new.keyword_parameters = keyword_parameters
        new.positional_parameters = positional_parameters
        return new



__all__ = ('WrapperConflict',)

from scarletio import RichAttributeErrorBaseType


class WrapperConflict(RichAttributeErrorBaseType):
    """
    Raised when 2 wrappers have conflict with each other.
    
    Attributes
    ----------
    wrapper_0 : ``WrapperBase``
        The wrapper 1 with the conflict.
    reason : `None`, `str`
        The reason why the two wrappers are incompatible.
    wrapper_1 : `None`, ``WrapperBase``
        The wrapper 2 with the conflict.
    """
    __slots__ = ('reason', 'wrapper_0', 'wrapper_1')
    
    def __new__(cls, wrapper_0, wrapper_1 = None, *, reason = None):
        """
        Creates a new wrapper conflict.
        
        Parameters
        ----------
        wrapper_0 : ``WrapperBase``
            Wrapper one with the conflict.
        wrapper_1 : `None`, ``WrapperBase`` = `None`, Optional
            Wrapper two with the conflict.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            The reason why the two wrappers are incompatible.
        """
        self = object.__new__(cls)
        self.wrapper_0 = wrapper_0
        self.wrapper_1 = wrapper_1
        self.reason = reason
        return self
    
    
    def __repr__(self):
        """Returns the wrapper conflict's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' wrapper_0 = ')
        repr_parts.append(repr(self.wrapper_0))
        
        wrapper_1 = self.wrapper_1
        if (wrapper_1 is not None):
            repr_parts.append(', wrapper_1 = ')
            repr_parts.append(repr(self.wrapper_1))
        
        reason = self.reason
        if (reason is not None):
            repr_parts.append(', reason = ')
            repr_parts.append(repr(self.reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the wrapper conflict's hash value."""
        hash_value = 0
        
        hash_value ^= hash(self.wrapper_0)
        
        wrapper_1 = self.wrapper_1
        if (wrapper_1 is not None):
            hash_value ^= hash(wrapper_1)
        
        reason = self.reason
        if (reason is not None):
            hash_value ^= hash(reason)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two wrapper conflicts are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.wrappers != other.wrappers:
            return False
        
        if self.reason != other.reason:
            return False
        
        return True
    
    
    @property
    def wrappers(self):
        """
        Returns the collective wrappers of the conflict.
        
        Returns
        -------
        wrappers : `frozenset` of ``WrapperBase``
        """
        wrappers = [self.wrapper_0]
        
        wrapper_1 = self.wrapper_1
        if (wrapper_1 is not None):
            wrappers.append(wrapper_1)
        
        return frozenset(wrappers)



__all__ = ()

from scarletio import DEFAULT_ANSI_HIGHLIGHTER, render_exception_into

from ...runner.runner import __file__ as VAMPYTEST_RUNNER_FILE_PATH


def ignore_invoke_test_frame(frame):
    """
    Ignores the frame where the import was called from.
    
    Parameters
    ----------
    frame : ``FrameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.frame
    name = frame.name
    line = frame.line
    
    if file_name == VAMPYTEST_RUNNER_FILE_PATH:
        if name == '_run_generator':
            if line == '__import__(source)':
                should_show_frame = False
    
    return should_show_frame


def render_load_failure_exception(exception):
    """
    Renders load failure exception
    
    Parameters
    ----------
    exception : ``BaseException``
        The raised exception.
    
    Returns
    -------
    rendered : `str`
    """
    return ''.join(
        render_exception_into(
            exception,
            [],
            filter = ignore_invoke_test_frame,
            highlighter = DEFAULT_ANSI_HIGHLIGHTER
        )
    )
